import ast
import os
import pickle
from attrs import define, field
from datetime import datetime
from typing import Any, Dict, Optional, Tuple, List, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import torch
from botorch.utils.multi_objective.hypervolume import Hypervolume
from botorch.utils.multi_objective.pareto import is_non_dominated
from gpytorch.kernels import AdditiveStructureKernel, Kernel, MaternKernel, ScaleKernel
from gpytorch.priors.torch_priors import GammaPrior

import nemo_bo.utils.logger as logging_nemo
import nemo_bo.utils.plotter as plotter
from nemo_bo.acquisition_functions.base_acq_function import AcquisitionFunction
from nemo_bo.acquisition_functions.expected_improvement.expected_improvement import ExpectedImprovement
from nemo_bo.acquisition_functions.highest_uncertainty.highest_uncertainty import HighestUncertainty
from nemo_bo.opt.benchmark import Benchmark, ModelBenchmark
from nemo_bo.opt.constraints import ConstraintsList
from nemo_bo.models.base.available_models import create_predictor_list
from nemo_bo.models.nn_bayesian import NNBayesianModel
from nemo_bo.models.nn_concrete import NNConcreteDropoutModel
from nemo_bo.opt.objectives import ObjectivesList, RegressionObjective
from nemo_bo.opt.samplers import LatinHyperCubeSampling, PoolBased, SampleGenerator
from nemo_bo.opt.variables import VariablesList
from nemo_bo.utils.data_proc import remove_nan, remove_all_nan_rows

try:
    logging_nemo.logging_path
    logger = logging_nemo.logging_nemo_child(os.path.basename(__file__))
except AttributeError:
    logger = logging_nemo.logging_nemo_master(os.path.basename(__file__))


@define
class Optimisation:
    """

    Class used to optimise a problem defined by the variables, objectives, acquisition function used to suggest new
    experiments, and optional constraints, sampler, and/or benchmark functions

    Parameters
    ----------
    variables: VariablesList
        VariablesList object that contains information about all variables
    objectives: ObjectivesList
        ObjectivesList object that contains information about all objectives
    acquisition_func: AcquisitionFunction
        Specifies the method used to suggest new candidates
    sampler: SampleGenerator, PoolBased, Default = None
        Specifies the method to obtain samples, either using a generator or from a sample pool
    constraints: ConstraintsList, Default = None
        ConstraintsList object that contains information about all constraints
    benchmark_func: Benchmark, Default = None
        Provides a method to predict the outcome of the suggested candidates made by the acquisition function when a
        simulated optimisation is being performed. If the benchmark provided is a ModelBenchmark object, please ensure
        that it has been fitted first
    opt_name: str, Default = None
        The name of the sub-folder to save the optimisation data inside

    """

    variables: VariablesList
    objectives: ObjectivesList
    acquisition_func: AcquisitionFunction
    sampler: Optional[Union[SampleGenerator, PoolBased]] = None
    constraints: Optional[ConstraintsList] = None
    benchmark_func: Optional[Benchmark] = None
    opt_name: Optional[str] = None
    optimisation_dict: Dict[str, Any] = field(init=False)
    X_columns: List[str] = field(init=False)
    Y_columns: List[str] = field(init=False)
    X: np.ndarray = field(init=False)
    Y: np.ndarray = field(init=False)
    optimisation_only: pd.DataFrame = field(init=False)
    selected_X: np.ndarray = field(init=False)
    selected_Y: np.ndarray = field(init=False)

    def __attrs_post_init__(self):
        if (
            isinstance(self.acquisition_func, ExpectedImprovement)
            or isinstance(self.acquisition_func, HighestUncertainty)
        ) and self.sampler is None:
            raise TypeError(
                "ExpectedImprovement or HighestUncertainty acquisition functions were passed as the acquisition "
                "function but no sampler was provided. Please pass a sampler too"
            )

        # Constructs lists of variables and objectives with their respective units where provided
        self.X_columns = []
        for name, units in zip(self.variables.names, self.variables.units):
            if units == "":
                self.X_columns.append(name)
            else:
                self.X_columns.append(f"{name} ({units})")

        self.Y_columns = []
        for name, units in zip(self.objectives.names, self.objectives.units):
            if units == "":
                self.Y_columns.append(name)
            else:
                self.Y_columns.append(f"{name} ({units})")

        # The name given to the optimisation run
        if self.opt_name is None:
            self.opt_name = f"Optimisation (start - {datetime.now().strftime('%d-%m-%Y, %H-%M-%S')})"

        # When a ModelBenchmark function has been provided, this checks that it has been fitted
        if self.benchmark_func is not None and isinstance(self.benchmark_func, ModelBenchmark):
            if not self.benchmark_func.fitted:
                raise AttributeError("Please firstly fit the benchmark function(s) and check its quality")
            else:
                self.benchmark_func.save(os.path.join(os.getcwd(), "Results", self.opt_name))

        # NNBayesianModel or NNConcreteDropoutModel models cannot be used with the ExpectedImprovement acquisition
        # function
        for obj in self.objectives.objectives:
            if (
                NNBayesianModel in obj.input_predictor_type or NNConcreteDropoutModel in obj.input_predictor_type
            ) and isinstance(self.acquisition_func, ExpectedImprovement):
                raise TypeError(
                    f"A NNBayesianModel or NNConcreteDropoutModel has been selected as a potential model "
                    f"for {obj.name} objective. However these should not be used with the "
                    f"ExpectedImprovement acquisition function due to extremely long run times. "
                    f"Either:\n1. Remove 'nn_bayesian' and 'nn_concrete' as options for the predictor type "
                    f"if you wish to continue using the ExpectedImprovement acquisition function\n2. Use "
                    f"the NSGAImprovement acquisition function instead if you wish to continue using these "
                    f"two types of models"
                )

    def run(
        self,
        X: Optional[np.ndarray] = None,
        Y: Optional[np.ndarray] = None,
        number_of_iterations: int = 100,
        test_ratio: float = 0.2,
        model_run_counter: int = 0,
        model_run_counter_threshold: int = 5,
        plot_progress: bool = False,
        plot_hypervolume: bool = False,
        test_data: Optional[List[np.ndarray]] = None,
        resume_run: bool = False,
        optimisation_pkl_path: Optional[str] = None,
        **kwargs,
    ) -> Dict[str, Dict[str, Any]]:
        """

        A convenient function to start a Bayesian optimisation, including model fitting. It will also request the results
        at the end of each iteration and save all optimisation data, and then automatically start the next iteration.

        Parameters
        ----------
        X: np.ndarray, Default = None,
            X array that contains the unprocessed variables data to be fitted to the model, i.e. The data has not yet
            undergone any transformations related to converted categorical variables to their respective descriptors
            or transformations related to standardisation/normalisation
        Y: np.ndarray, Default = None,
            Y array that contains the unprocessed objective data for all calculable and regression objectives, i.e.
            The data has not yet undergone any transformations related to standardisation/normalisation
        number_of_iterations: int, Default = 100,
            Number of optimisation iterations to perform
        test_ratio: float, Default = 0.2,
            The proportion of inputted X and Y arrays to be split for the validation and test sets where applicable
        model_run_counter: int, Default = 0,
            Assigns the run counter for the first iteration when this function is called. When the model_run_counter is
            equal to model_run_counter_threshold, an automated model and hyperparameter optimisation will occur at the
            next opportunity
        model_run_counter_threshold: int, Default = 5,
            Sets the threshold value for when an automated model and hyperparameter optimisation should occur
        plot_progress: bool, Default = False,
            Sets whether to generate an objective values over iterations scatter plot for 1 objective optimisations or
            a 2D/3D pareto plot (2/3 objectives) every iteration
        plot_hypervolume: bool, Default = False,
            Sets whether to generate a multiobjective hypervolume over iterations plot every iteration
        test_data: List[np.ndarray], Default = None,
            Test data to use for the initial model and hyperparameter optimisation, instead of creating it
            automatically from the provided X and Y arrays. The test X array is at index 0 and the test Y array is at
            index 1. Note: the test data are appended to the X and Y arrays for the final model fitting (after the best
            model type has been identified) and the acquisition function
        resume_run: bool, Default = False,
            Sets whether to resume an optimisation using NEMO. Ensure that optimisation_pkl_path is passed too
        optimisation_pkl_path: str, Default = None,
            String of the file path of the pickle file written during the previous NEMO optimisation run

        Returns
        -------
        self.optimisation_dict: Dict[str, Dict[str, Any]]
            Dictionary that stores the optimisation information for every iteration

        """
        # if number_opt_completed is None:
        if not resume_run:
            number_opt_completed = []

            # Dictionary that stores the optimisation information for every iteration
            self.optimisation_dict = {}

        # If resuming a previous optimisation
        if resume_run and optimisation_pkl_path is not None:
            # Reads the optimisation pickle file at optimisation_pkl_path
            with open(optimisation_pkl_path, "rb") as file:
                self.optimisation_dict = pickle.load(file)

            # Loads the information from the most recent iteration
            last_dict = self.optimisation_dict[list(self.optimisation_dict.keys())[-1]]

            # Extracts the X and Y arrays from the optimisation run so far
            XY = last_dict["Training + Optimisation Table"]
            X = XY.to_numpy()[:, : (-1 * self.objectives.n_obj)]
            Y = XY.to_numpy()[:, (-1 * self.objectives.n_obj) :].astype(float)

            # Loads the optimisation run information
            self.optimisation_only = last_dict["Optimisation Table"]
            model_run_counter = int(self.optimisation_only["Model Run Counter"].iloc[-1])

            # Creates number_opt_completed list
            number_opt_completed = [
                (
                    self.optimisation_only["Number of Candidates in Iteration"].astype("int")[
                        self.optimisation_only["Iteration Number"].astype("int") == i
                    ]
                ).iloc[0]
                for i in range(1, int(self.optimisation_only["Iteration Number"].iloc[-1]) + 1)
            ]

            # Assigns the model types and hyperparameters to the objectives
            for obj_index, (obj_name, obj) in enumerate(zip(self.objectives.names, self.objectives.objectives)):
                if isinstance(obj, RegressionObjective):
                    self.objectives.objectives[obj_index].predictor_type = create_predictor_list(
                        last_dict[f"{obj_name} Model Type"]
                    )[0]

                    if last_dict[f"{obj_name} Model Type"] == "gp":
                        self.objectives.objectives[obj_index].model_params = {
                            "kernel": eval(last_dict[f"{obj_name} Model Hyperparameters"])
                        }
                    else:
                        self.objectives.objectives[obj_index].model_params = last_dict[
                            f"{obj_name} Model Hyperparameters"
                        ]

                    self.objectives.objectives[obj_index].obj_function = self.objectives.objectives[
                        obj_index
                    ].predictor_type(self.variables, self.objectives.objectives[obj_index])

        # If no X and Y arrays are provided
        if X is None and Y is None:
            # Only valid to supply no X and Y arrays when a benchmark function is also provided
            if self.benchmark_func is None:
                raise ValueError(
                    f"No X and Y arrays were provided. If you wish to simulate the outcome of experiments "
                    f"in a closed-loop manner, please provide a benchmark function and then a set of X and "
                    f"Y values will be automatically generated. If you wish to carry out actual Bayesian "
                    f"optimisation experiments, please pass X and Y arrays"
                )
            else:
                # Generates an initial training set using the supplied sampler and benchmark function
                X = self.sampler.__class__(
                    (2 * self.variables.n_var) + 2 if self.variables.n_var >= 4 else 10
                ).generate_samples(self.variables, self.constraints)
                Y, _ = self.benchmark_func.evaluate(X)
                if Y.ndim == 1:
                    Y = Y.reshape(-1, 1)

        if self.objectives.n_obj == 1 and Y.ndim == 1:
            Y = Y.reshape(-1, 1)

        self.X = X
        self.Y = Y

        # (Re-)Starts the optimisation run
        for iteration in range(len(number_opt_completed), number_of_iterations):
            # Checks if automated model and hyperparameter selection will be performed in this iteration
            if model_run_counter % model_run_counter_threshold == 0 or model_run_counter > model_run_counter_threshold:
                model_search_bool = True
                logger.info("Will perform automated model and hyperparameter selection")

            else:
                model_search_bool = False
                logger.info("Will perform model fitting using the previously identified best model type")

            if plot_progress:
                # Scatter plot of the Y data
                self.plot_scatter_opt(
                    Y=self.Y,
                    number_opt_completed=number_opt_completed,
                    best=self.benchmark_func.best_result() if self.benchmark_func is not None else None,
                )

            # Plot hypervolume over iterations plot
            if plot_hypervolume:
                if self.objectives.n_obj > 1:
                    self.plot_hypervolume(
                        Y=self.Y,
                        number_opt_completed=number_opt_completed,
                    )
                else:
                    raise AttributeError(
                        f"The number of objectives ({self.objectives.n_obj}) needs to be 2 or more to "
                        f"create a hypervolume plot"
                    )

            # Obtain suggested candidates
            self.selected_X, self.selected_Y = self.find_candidates(
                self.X,
                self.Y,
                model_search_bool=model_search_bool,
                test_ratio=test_ratio,
                test_data=test_data,
                **kwargs,
            )
            if self.selected_Y.ndim == 1:
                self.selected_Y = self.selected_Y.reshape(-1, 1)

            # Creates a DataFrame of the suggested candidates
            suggested_XY = pd.DataFrame(
                np.hstack((self.selected_X, self.selected_Y)),
                columns=self.X_columns + self.Y_columns,
            )
            with pd.option_context("display.max_rows", None, "display.max_columns", None):
                print(suggested_XY)

            # If the optimisation is not a simulated one that used a sample pool or a benchmark function
            if not isinstance(self.sampler, PoolBased) and self.benchmark_func is None:
                if not os.path.exists(os.path.join(os.getcwd(), "Results")):
                    os.makedirs(os.path.join(os.getcwd(), "Results"))
                # Creates an Excel file containing the suggested candidates
                suggested_XY.to_excel(
                    os.path.join(
                        os.getcwd(),
                        "Results",
                        self.opt_name,
                        f"Iteration {iteration + 1} suggested candidates "
                        f"{datetime.now().strftime('%d-%m-%Y, %H-%M-%S')}.xlsx",
                    )
                )
                while True:
                    try:
                        # Requests a new array containing the actual Y-values. The array has to be the correct shape
                        # and a list type to be accepted
                        actual_result = ast.literal_eval(
                            input(
                                f"Variables for the current opt iteration have been suggested (see print or "
                                f"'Iteration {iteration + 1} suggested candidates.xlsx')\nPlease input the actual "
                                f"results in a python list:\n"
                            )
                        )
                        assert isinstance(actual_result, list)

                        self.selected_Y = (
                            np.array(actual_result)
                            if self.acquisition_func.num_candidates > 1
                            else np.array(actual_result).reshape(1, -1)
                        )

                        assert (
                            self.selected_Y.shape[0] == self.acquisition_func.num_candidates
                            and self.selected_Y.shape[1] == self.objectives.n_obj
                            and np.issubdtype(self.selected_Y.dtype, np.number)
                        )

                        break
                    except AssertionError or IndexError or ValueError or SyntaxError:
                        print(
                            "\nAssertionError: Please ensure that you have inputted the actual numerical results in a "
                            "python list with the correct dimensions for the number of candidates (x) and number of "
                            "objectives (y)\n"
                        )

            # Updates the X and Y arrays with the new candidates
            self.X = np.vstack((self.X, self.selected_X))
            self.Y = np.vstack((self.Y, self.selected_Y))

            # After every iteration, adds 1 to the model_run_counter number
            if model_search_bool:
                model_run_counter = 1
            else:
                model_run_counter += 1

            # Saves the optimisation and model information for the current iteration
            self.save(iteration + 1, model_search_bool, model_run_counter)

            number_opt_completed.append(self.acquisition_func.num_candidates)

        # When all iterations ahve been completed, the final plots will be generated
        if plot_progress:
            # Scatter plot of the Y data
            self.plot_scatter_opt(
                Y=self.Y,
                number_opt_completed=number_opt_completed,
                best=self.benchmark_func.best_result() if self.benchmark_func is not None else None,
            )

        # Plot hypervolume over iterations plot
        if plot_hypervolume:
            if self.objectives.n_obj > 1:
                self.plot_hypervolume(
                    Y=self.Y,
                    number_opt_completed=number_opt_completed,
                )
            else:
                raise AttributeError(
                    f"The number of objectives ({self.objectives.n_obj}) needs to be 2 or more to "
                    f"create a hypervolume plot"
                )

        return self.optimisation_dict

    def find_candidates(
        self,
        X: np.ndarray,
        Y: np.ndarray,
        model_search_bool,
        test_ratio: float = 0.2,
        test_data: Optional[List[np.ndarray]] = None,
        **kwargs,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """

        Function that returns the suggested candidates by:
        1. Fitting any required regression models
        2. Uses an acquisition function to suggest the next best candidates
        3. Updates the Y-values with

        Parameters
        ----------
        X: np.ndarray,
            X array that contains the unprocessed variables data to be fitted to the model, i.e. The data has not yet
            undergone any transformations related to converted categorical variables to their respective descriptors
            or transformations related to standardisation/normalisation
        Y: np.ndarray,
            Y array that contains the unprocessed objective data for all calculable and regression objectives, i.e.
            The data has not yet undergone any transformations related to standardisation/normalisation
        model_search_bool: bool
            Whether automated model and hyperparameter optimisation is to be performed when fitting the regression
            models
        test_ratio: float, Default = 0.2,
            The proportion of inputted X and Y arrays to be split for the validation and test sets where applicable
        test_data: List[np.ndarray], Default = None,
            Test data to use for the initial model and hyperparameter optimisation, instead of creating it
            automatically from the provided X and Y arrays. The test X array is at index 0 and the test Y array is at
            index 1. Note: the test data are appended to the X and Y arrays for the final model fitting (after the best
            model type has been identified) and the acquisition function

        Returns
        -------
        selected_X: np.ndarray
            The X array of the suggested candidates
        selected_Y: np.ndarray
            The Y array of the suggested candidates

        """
        if any(isinstance(obj, RegressionObjective) for obj in self.objectives.objectives):
            # Fit the ML regression models
            self.objectives.fit(
                X, Y, self.variables, model_search_bool, test_ratio=test_ratio, test_data=test_data, **kwargs
            )

        # Use the fitted regression models, calculable objectives, and acquisition function to suggest new candidates.
        # If test data was supplied, it is appended to X and Y for the acquisition function
        if test_data is not None:
            X = np.vstack((X, test_data[0]))
            Y = np.vstack((Y, test_data[1]))

        if isinstance(self.acquisition_func, ExpectedImprovement):
            selected_X, selected_Y = self.acquisition_func.generate_candidates(
                X, Y, self.variables, self.objectives, self.sampler, self.constraints, **kwargs
            )

        else:
            selected_X, selected_Y = self.acquisition_func.generate_candidates(
                Y, self.variables, self.objectives, sampler=self.sampler, constraints=self.constraints, **kwargs
            )

        # If a sample pool was used, obtain the actual Y-values and removing the chosen samples from the sample pool
        if isinstance(self.sampler, PoolBased):
            selected_Y = self.sampler.Y_and_update_pool(self.sampler.index)

        else:
            # If a benchmark function is provided, use it to predict the outcome of the suggested X array
            if self.benchmark_func is not None:
                if self.benchmark_func.fitted:
                    selected_Y, _ = self.benchmark_func.evaluate(selected_X)
                else:
                    raise AttributeError("Please firstly fit the benchmark function(s) and check its quality")

        return selected_X, selected_Y

    def save(self, iteration: int, model_search_bool: bool, model_run_counter: int) -> None:
        """

        Saves the optimisation and model information for the current iteration

        Parameters
        ----------
        iteration: int
            Iteration number for the current optimisation
        model_search_bool: bool
            Whether automated model and hyperparameter optimisation was performed when the regression models were fitted
        model_run_counter: int
            The run counter for this optimisation iteration

        """
        if not os.path.exists(os.path.join(os.getcwd(), "Results", self.opt_name)):
            os.makedirs(os.path.join(os.getcwd(), "Results", self.opt_name))

        # Temporarily store the information in these lists
        model_info_list = [iteration, self.acquisition_func.num_candidates, model_search_bool, model_run_counter]
        model_info_columns = [
            "Iteration Number",
            "Number of Candidates in Iteration",
            "Model and Hyperparameter Optimisation",
            "Model Run Counter",
        ]

        if self.objectives.n_obj > 1:
            model_info_list.append(self.calc_hypervolume(self.Y))
            model_info_columns.append("Normalised Hypervolume")

        for obj in self.objectives.objectives:
            if isinstance(obj, RegressionObjective):
                if obj.obj_function.include_validation:
                    model_info_list.append(obj.obj_function.performance_metrics["Train RMSE"])
                    model_info_list.append(obj.obj_function.performance_metrics["Train r2"])
                    model_info_list.append(obj.obj_function.performance_metrics["Validation RMSE"])
                    model_info_list.append(obj.obj_function.performance_metrics["Validation r2"])
                    model_info_columns.append(f"{obj.name} Model Train RMSE")
                    model_info_columns.append(f"{obj.name} Model Train r2")
                    model_info_columns.append(f"{obj.name} Model Validation RMSE")
                    model_info_columns.append(f"{obj.name} Model Validation r2")
                else:
                    model_info_list.append(obj.obj_function.performance_metrics["RMSE"])
                    model_info_list.append(obj.obj_function.performance_metrics["r2"])
                    model_info_columns.append(f"{obj.name} Model Train RMSE")
                    model_info_columns.append(f"{obj.name} Model Train r2")
                model_info_list.append(obj.obj_function.name)

                if obj.obj_function.name == "gp":
                    if isinstance(obj.model_params["kernel"], AdditiveStructureKernel):
                        model_info_list.append(
                            f"AdditiveStructureKernel(ScaleKernel(MaternKernel(ard_num_dims={self.variables.n_var}, "
                            f"nu={obj.model_params['kernel'].base_kernel.base_kernel.nu}, "
                            f"lengthscale_prior=GammaPrior(3.0, 6.0)), outputscale_prior=GammaPrior(2.0, 0.15)), "
                            f"num_dims={self.variables.n_var})"
                        )
                    else:
                        model_info_list.append(
                            f"ScaleKernel(MaternKernel(nu={obj.model_params['kernel'].base_kernel.nu}, "
                            f"ard_num_dims={self.variables.n_var}, lengthscale_prior=GammaPrior(3.0, 6.0)), "
                            f"outputscale_prior=GammaPrior(2.0, 0.15))"
                        )
                else:
                    model_info_list.append(obj.model_params)

                model_info_columns.append(f"{obj.name} Model Type")
                model_info_columns.append(f"{obj.name} Model Hyperparameters")

        model_info_list.append(datetime.now().strftime("%d-%m-%Y, %H-%M-%S"))
        model_info_columns.append("Date and Time")

        # Duplicates this information for every candidate
        model_info = np.array([model_info_list] * self.acquisition_func.num_candidates)

        # Creates the DataFrames of optimisation information
        optimisation_results = pd.DataFrame(np.hstack((self.X, self.Y)), columns=self.X_columns + self.Y_columns)
        if iteration == 1:
            self.optimisation_only = pd.DataFrame(
                np.hstack((self.selected_X, self.selected_Y, model_info)),
                columns=self.X_columns + self.Y_columns + model_info_columns,
            )
        else:
            self.optimisation_only = pd.concat(
                (
                    self.optimisation_only,
                    pd.DataFrame(
                        np.hstack((self.selected_X, self.selected_Y, model_info)),
                        columns=self.X_columns + self.Y_columns + model_info_columns,
                    ),
                ),
                axis=0,
                ignore_index=True,
            )

        # Writes the optimisation information DataFrames into Excel files
        with pd.ExcelWriter(
            os.path.join(
                os.getcwd(),
                "Results",
                self.opt_name,
                f"Optimisation Results (after {iteration} iteration(s)) "
                f'{datetime.now().strftime("%d-%m-%Y, %H-%M-%S")}.xlsx',
            )
        ) as writer:
            optimisation_results.to_excel(writer, sheet_name="Training + Opt")
            self.optimisation_only.to_excel(writer, sheet_name="Only Opt")
            for obj in self.objectives.objectives:
                if isinstance(obj, RegressionObjective):
                    if obj.obj_function.include_validation:
                        obj.obj_function.performance_metrics["Train Parity Data"].to_excel(
                            writer, sheet_name=f"{obj.name} Train Parity Data"
                        )
                        obj.obj_function.performance_metrics["Validation Parity Data"].to_excel(
                            writer, sheet_name=f"{obj.name} Val Parity Data"
                        )
                    else:
                        obj.obj_function.performance_metrics["Parity Data"].to_excel(
                            writer, sheet_name=f"{obj.name} Train Parity Data"
                        )

        # Writes pickle files containing every iteration's optimisation information
        self.optimisation_dict[f"Iteration {iteration}"] = {
            "Date and Time": datetime.now().strftime("%d-%m-%Y, %H-%M-%S"),
            "Training + Optimisation Table": optimisation_results,
            "Optimisation Table": self.optimisation_only,
            "Optimisation Selected X": self.selected_X,
            "Optimisation Selected Y": self.selected_Y,
            "Model and Hyperparameter Optimisation": model_search_bool,
            "Model Run Counter": model_run_counter,
        }
        for obj in self.objectives.objectives:
            if isinstance(obj, RegressionObjective):
                self.optimisation_dict[f"Iteration {iteration}"][f"{obj.name} Model Type"] = obj.obj_function.name
                if obj.obj_function.name == "gp":
                    if isinstance(obj.model_params["kernel"], AdditiveStructureKernel):
                        self.optimisation_dict[f"Iteration {iteration}"][f"{obj.name} Model Hyperparameters"] = (
                            f"AdditiveStructureKernel(ScaleKernel(MaternKernel(ard_num_dims={self.variables.n_var}, "
                            f"nu={obj.model_params['kernel'].base_kernel.base_kernel.nu}, "
                            f"lengthscale_prior=GammaPrior(3.0, 6.0)), outputscale_prior=GammaPrior(2.0, 0.15)), "
                            f"num_dims={self.variables.n_var})"
                        )
                    else:
                        self.optimisation_dict[f"Iteration {iteration}"][f"{obj.name} Model Hyperparameters"] = (
                            f"ScaleKernel(MaternKernel(nu={obj.model_params['kernel'].base_kernel.nu}, "
                            f"ard_num_dims={self.variables.n_var}, lengthscale_prior=GammaPrior(3.0, 6.0)), "
                            f"outputscale_prior=GammaPrior(2.0, 0.15))"
                        )
                else:
                    self.optimisation_dict[f"Iteration {iteration}"][
                        f"{obj.name} Model Hyperparameters"
                    ] = obj.model_params
                self.optimisation_dict[f"Iteration {iteration}"][
                    f"{obj.name} Model Performance Metrics"
                ] = obj.obj_function.performance_metrics

        with open(
            os.path.join(
                os.getcwd(),
                "Results",
                self.opt_name,
                f"Optimisation Results (after {iteration} iteration(s)) "
                f"{datetime.now().strftime('%d-%m-%Y, %H-%M-%S')}.pickle",
            ),
            "wb",
        ) as file:
            pickle.dump(self.optimisation_dict, file, protocol=pickle.HIGHEST_PROTOCOL)

    def calc_hypervolume(self, Y: np.ndarray) -> float:
        """

        Calculates the hypervolume of a Y-array that has been min-max normalised by the pre-defined objective bounds

        Parameters
        ----------
        Y: np.ndarray
            Y-array with objective data. Any data points with NaN values will be removed

        Returns
        -------
        The normalised hypervolume of a Y-array

        """
        Y_no_nan = remove_all_nan_rows(Y)
        ref_point = self.acquisition_func.build_ref_point(self.objectives.max_bool_dict)
        sign_adjusted_Y = self.acquisition_func.Y_norm_minmax_transform(
            Y_no_nan, self.objectives.bounds, self.objectives.max_bool_dict
        )

        # Calculates pareto front
        pareto_mask = is_non_dominated(torch.tensor(sign_adjusted_Y, dtype=torch.double))
        pareto_front = sign_adjusted_Y[pareto_mask]

        # Calculates the hypervolume
        hv = Hypervolume(ref_point=torch.tensor(ref_point, dtype=torch.double))
        return hv.compute(torch.tensor(pareto_front, dtype=torch.double))

    def plot_scatter_opt(
        self, Y: np.ndarray, number_opt_completed: List[int], best: Optional[Union[int, float, np.ndarray]] = None
    ) -> None:
        """

        Plots the Y-array for 1, 2, or 3 objective optimisation problems, with multi-objective pareto fronts when
        applicable

        Parameters
        ----------
        Y: np.ndarray
            Y-array with objective data. Any data points with NaN values will be removed
        number_opt_completed: List[int]
            List containing the number of optimisation candidates completed per iteration, e.g. [2, 4] means
            2 candidates were run in the first iteration and 4 candidates in the second iteration
        best: [int | float | np.ndarray], Default = None
            The actual best result. Would be a single value for a single objective problem or a np.ndarray for a
            multiobjective problem

        """
        if not os.path.exists(os.path.join(os.getcwd(), "Results", self.opt_name)):
            os.makedirs(os.path.join(os.getcwd(), "Results", self.opt_name))

        # Calculates the number of training and optimisation experiments
        opt_num = sum(number_opt_completed)
        train_num = Y.shape[0] - opt_num

        if self.objectives.n_obj == 1:
            iterations = np.array((range(Y.shape[0]))) + 1
            Y = np.hstack((iterations.reshape(-1, 1), Y))

        else:
            Y_no_nan = remove_all_nan_rows(Y)

            sign_adjusted_Y = self.acquisition_func.Y_norm_minmax_transform(
                Y_no_nan, self.objectives.bounds, self.objectives.max_bool_dict
            )

            # Calculates pareto front
            Y_to_expnum = torch.tensor(np.array(sign_adjusted_Y), dtype=torch.double)
            pareto_mask_to_expnum = is_non_dominated(Y_to_expnum)
            pareto_front = Y_no_nan[pareto_mask_to_expnum]

            non_zero_objs = np.where(np.all(~np.isclose(Y_no_nan, 0.0), axis=0))[0]
            if len(non_zero_objs) > 0:
                # Uses the first objective in the array that has all non-zero values
                pareto_front_sorted = pareto_front[pareto_front[:, non_zero_objs[0]].argsort()]
            else:
                # Else just use the first objective
                pareto_front_sorted = pareto_front[pareto_front[:, 0].argsort()]

        train_experiments = Y[
            :train_num,
            :,
        ]
        opt_experiments = Y[
            train_num:,
            :,
        ]

        train_experiments = remove_all_nan_rows(train_experiments)
        opt_experiments = remove_all_nan_rows(opt_experiments)

        scatter_data = [train_experiments, opt_experiments]
        scatter_legend = [
            f"Training ({train_num} experiments)"
            if train_num == train_experiments.shape[0]
            else f"Training ({train_num} expt, showing {train_experiments.shape[0]} expt)",
            f"Optimisation ({opt_num} experiments)"
            if opt_num == opt_experiments.shape[0]
            else f"Optimisation ({opt_num} expt, showing {opt_experiments.shape[0]} expt)",
        ]

        output = os.path.join(
            os.getcwd(),
            "Results",
            self.opt_name,
            f'Pareto Plot (after {len(number_opt_completed)} iteration(s)) {datetime.now().strftime("%d-%m-%Y, %H-%M-%S")}'
            if self.objectives.n_obj > 1
            else f'Progress Plot (after {len(number_opt_completed)} iteration(s)) {datetime.now().strftime("%d-%m-%Y, %H-%M-%S")}',
        )

        # If single objective
        if self.objectives.n_obj == 1:
            opt_line = [[train_num + 0.5, min(Y[:, 1])], [train_num + 0.5, max(Y[:, 1])]]
            best_line = [[0, best], [Y.shape[0], best]]
            line_data = [opt_line, best_line] if best is not None else opt_line

            plotter.plot(
                plot_dim="2D",
                scatter_data=scatter_data,
                # Create array to use for drawing the start of the optimisation line
                line_data=line_data,
                scatter_legend=scatter_legend,
                line_legend=["Start of optimisation", "Best possible result"]
                if best is not None
                else "Start of optimisation",
                xlabel=f"Number of experiments",
                ylabel=f"{self.objectives.names[0]} ({self.objectives.units[0]})"
                if self.objectives.units[0] != ""
                else f"{self.objectives.names[0]}",
                plottitle=f"Progress Plot\n{self.objectives.names[0]} vs. Number of experiments",
                output_file=rf"{output}",
            )
        # If two objectives
        elif self.objectives.n_obj == 2:
            plotter.plot(
                plot_dim="2D",
                scatter_data=scatter_data + [best] if best is not None else scatter_data,
                line_data=pareto_front_sorted,
                scatter_legend=scatter_legend + ["Approx. best pareto front"] if best is not None else scatter_legend,
                line_legend="Observed pareto front",
                xlabel=f"{self.objectives.names[0]} ({self.objectives.units[0]})"
                if self.objectives.units[0] != ""
                else f"{self.objectives.names[0]}",
                ylabel=f"{self.objectives.names[1]} ({self.objectives.units[1]})"
                if self.objectives.units[1] != ""
                else f"{self.objectives.names[1]}",
                plottitle=f"Pareto Plot\n{self.objectives.names[0]} vs. {self.objectives.names[1]}",
                output_file=rf"{output}",
            )

        # If three objectives
        elif self.objectives.n_obj == 3:
            plotter.plot(
                plot_dim="3D",
                scatter_data=scatter_data,
                surface_data=[pareto_front_sorted, best],
                scatter_legend=scatter_legend,
                surface_legend="Pareto front",
                xlabel=f"{self.objectives.names[0]} ({self.objectives.units[0]})"
                if self.objectives.units[0] != ""
                else f"{self.objectives.names[0]}",
                ylabel=f"{self.objectives.names[1]} ({self.objectives.units[1]})"
                if self.objectives.units[1] != ""
                else f"{self.objectives.names[1]}",
                zlabel=f"{self.objectives.names[2]} ({self.objectives.units[2]})"
                if self.objectives.units[2] != ""
                else f"{self.objectives.names[2]}",
                plottitle=f"Pareto Plot\n{self.objectives.names[0]} vs. {self.objectives.names[1]} vs. {self.objectives.names[2]}",
                output_file=rf"{output}",
            )

        plt.close()

    def plot_hypervolume(self, Y: np.ndarray, number_opt_completed: List[int]) -> None:
        """

        Plots the hypervolume over all iterations

        Parameters
        ----------
        Y: np.ndarray
            Y-array with objective data. Any data points with NaN values will be removed
        number_opt_completed: List[int]
            List containing the number of optimisation candidates completed per iteration, e.g. [2, 4] means
            2 candidates were run in the first iteration and 4 candidates in the second iteration

        """
        if not os.path.exists(os.path.join(os.getcwd(), "Results", self.opt_name)):
            os.makedirs(os.path.join(os.getcwd(), "Results", self.opt_name))

        # Calculates the number of training and optimisation experiments
        opt_num = sum(number_opt_completed)
        train_num = Y.shape[0] - opt_num

        iterations = np.array((range(Y.shape[0]))) + 1
        del_index = [index for index in range(Y.shape[0]) if np.any(np.isnan(Y[index]), axis=0)]
        iterations = np.delete(iterations, del_index, axis=0)

        no_nan_train_num = sum(expt_number <= train_num for expt_number in iterations)
        no_nan_opt_num = sum(expt_number > train_num for expt_number in iterations)

        Y_no_nan = remove_all_nan_rows(Y)
        ref_point = self.acquisition_func.build_ref_point(self.objectives.max_bool_dict)
        sign_adjusted_Y = self.acquisition_func.Y_norm_minmax_transform(
            Y_no_nan, self.objectives.bounds, self.objectives.max_bool_dict
        )
        # Recalculates the hypervolume for every additional row in the data
        hv_list = [0]
        for y_index in range(2, Y_no_nan.shape[0] + 1):
            # Calculates pareto front
            obj_data_to_expnum = sign_adjusted_Y[:y_index, :]
            obj_data_to_expnum_np = np.array(obj_data_to_expnum)
            Y_to_expnum = torch.tensor(obj_data_to_expnum_np, dtype=torch.double)
            pareto_mask_to_expnum = is_non_dominated(Y_to_expnum)
            pareto_front_to_expnum = obj_data_to_expnum_np[pareto_mask_to_expnum]

            # Calculates the hypervolume
            hv = Hypervolume(ref_point=torch.tensor(ref_point, dtype=torch.double))
            hv_list.append(hv.compute(torch.tensor(pareto_front_to_expnum, dtype=torch.double)))

        if Y.shape[0] == len(iterations):
            title = f"Hypervolume\nTraining: {train_num} expt\nOptimisation: {opt_num} expt"
        else:
            title = (
                f"Hypervolume\nTraining: {train_num} expt, showing {no_nan_train_num} expt\n"
                f"Optimisation: {opt_num} expt, showing {no_nan_opt_num} expt"
            )

        output_file = os.path.join(
            os.getcwd(),
            "Results",
            self.opt_name,
            f"Hypervolume (after {len(number_opt_completed)} iteration(s)) {datetime.now().strftime('%d-%m-%Y, %H-%M-%S')}",
        )

        hv_array = np.array([[expt_number, hv_list[index]] for index, expt_number in enumerate(iterations)])
        plotter.plot(
            plot_dim="2D",
            scatter_data=hv_array,
            line_data=[[train_num + 0.5, 0.0], [train_num + 0.5, 1.0]],
            scatter_legend="Hypervolume",
            line_legend="Start of Optimisation",
            xlabel="Number of experiments",
            ylabel="Normalised Hypervolume",
            plottitle=title,
            output_file=output_file,
        )
        plt.close()



# This sampling code was copied from the smt package by M. A. Bouhlel and J. T. Hwang and N. Bartoli and R. Lafage and
# J. Morlier and J. R. R. A. Martins, A Python surrogate modeling framework with derivatives. Advances in Engineering
# Software, 2019, 102662
# https://doi.org/10.1016/j.advengsoft.2019.03.005
# https://smt.readthedocs.io/en/latest/

from abc import ABCMeta, abstractmethod

import numpy as np
from pyDOE2 import lhs
from scipy.spatial.distance import pdist, cdist

FLOAT = "float_type"
INT = "int_type"
ORD = "ord_type"
ENUM = "enum_type"

#######################################


def ensure_2d_array(array, name):
    if not isinstance(array, np.ndarray):
        raise ValueError("{} must be a NumPy array".format(name))

    array = np.atleast_2d(array.T).T

    if len(array.shape) != 2:
        raise ValueError("{} must have a rank of 1 or 2".format(name))

    return array


def check_support(sm, name, fail=False):
    if not sm.supports[name] or fail:
        class_name = sm.__class__.__name__
        raise NotImplementedError("{} does not support {}".format(class_name, name))


def check_nx(nx, x):
    if x.shape[1] != nx:
        if nx == 1:
            raise ValueError("x should have shape [:, 1] or [:]")
        else:
            raise ValueError("x should have shape [:, {}] and not {}".format(nx, x.shape))


#######################################


def bisect_left(*args, **kwargs):  # real signature unknown
    """
    Return the index where to insert item x in list a, assuming a is sorted.

    The return value i is such that all e in a[:i] have e < x, and all e in
    a[i:] have e >= x.  So if x already appears in the list, i points just
    before the leftmost x already there.

    Optional args lo (default 0) and hi (default len(a)) bound the
    slice of a to be searched.
    """
    pass


#######################################


def take_closest_number(myList, myNumber):
    """
    Assumes myList is sorted. Returns closest value to myNumber.

    If two numbers are equally close, return the smallest number.
    """
    pos = bisect_left(myList, myNumber)
    if pos == 0:
        return myList[0]
    if pos == len(myList):
        return myList[-1]
    before = myList[pos - 1]
    after = myList[pos]
    if after - myNumber < myNumber - before:
        return after
    else:
        return before


def take_closest_in_list(myList, x):
    vfunc = np.vectorize(take_closest_number, excluded=["myList"])
    return vfunc(myList=myList, myNumber=x)


#######################################


class OptionsDictionary(object):
    """
    Generalization of the dictionary that allows for declaring keys.

    Attributes
    ----------
    _dict : dict
        Dictionary of option values keyed by option names.
    _declared_entries : dict
        Dictionary of declared entries.
    """

    def __init__(self):
        self._dict = {}
        self._declared_entries = {}

    def clone(self):
        """
        Return a clone of this object.

        Returns
        -------
        OptionsDictionary
            Deep-copied clone.
        """
        clone = self.__class__()
        clone._dict = dict(self._dict)
        clone._declared_entries = dict(self._declared_entries)
        return clone

    def __getitem__(self, name):
        """
        Get an option that was previously declared and optionally set.

        Arguments
        ---------
        name : str
            The name of the option.

        Returns
        -------
        object
            Value of the option.
        """
        return self._dict[name]

    def __setitem__(self, name, value):
        """
        Set an option that was previously declared.

        The value argument must be valid, which means it must satisfy the following:
        1. If values and not types was given when declaring, value must be in values.
        2. If types and not values was given when declaring, type(value) must be in types.
        3. If values and types were given when declaring, either of the above must be true.

        Arguments
        ---------
        name : str
            The name of the option.
        value : object
            The value to set.
        """
        assert name in self._declared_entries, "Option %s has not been declared" % name
        self._assert_valid(name, value)
        self._dict[name] = value

    def __contains__(self, key):
        return key in self._dict

    def is_declared(self, key):
        return key in self._declared_entries

    def _assert_valid(self, name, value):
        values = self._declared_entries[name]["values"]
        types = self._declared_entries[name]["types"]

        if values is not None and types is not None:
            assert value in values or isinstance(
                value, types
            ), "Option %s: value and type of %s are both invalid - " % (
                name,
                value,
            ) + "value must be %s or type must be %s" % (
                values,
                types,
            )
        elif values is not None:
            assert value in values, "Option %s: value %s is invalid - must be %s" % (
                name,
                value,
                values,
            )
        elif types is not None:
            assert isinstance(value, types), "Option %s: type of %s is invalid - must be %s" % (name, value, types)

    def update(self, dict_):
        """
        Loop over and set all the entries in the given dictionary into self.

        Arguments
        ---------
        dict_ : dict
            The given dictionary. All keys must have been declared.
        """
        for name in dict_:
            self[name] = dict_[name]

    def declare(self, name, default=None, values=None, types=None, desc=""):
        """
        Declare an option.

        The value of the option must satisfy the following:
        1. If values and not types was given when declaring, value must be in values.
        2. If types and not values was given when declaring, type(value) must be in types.
        3. If values and types were given when declaring, either of the above must be true.

        Arguments
        ---------
        name : str
            Name of the option.
        default : object
            Optional default value that must be valid under the above 3 conditions.
        values : list
            Optional list of acceptable option values.
        types : type or list of types
            Optional list of acceptable option types.
        desc : str
            Optional description of the option.
        """
        self._declared_entries[name] = {
            "values": values,
            "types": types,
            "default": default,
            "desc": desc,
        }

        if default is not None:
            self._assert_valid(name, default)

        self._dict[name] = default


#######################################


class SamplingMethod(metaclass=ABCMeta):
    def __init__(self, **kwargs):
        """
        Constructor where values of options can be passed in.

        For the list of options, see the documentation for the problem being used.

        Parameters
        ----------
        **kwargs : named arguments
            Set of options that can be optionally set; each option must have been declared.

        Examples
        --------
        >>> import numpy as np
        >>> from smt.sampling_methods import Random
        >>> sampling = Random(xlimits=np.arange(2).reshape((1, 2)))
        """
        self.options = OptionsDictionary()
        self.options.declare(
            "xlimits",
            types=np.ndarray,
            desc="The interval of the domain in each dimension with shape nx x 2 (required)",
        )
        self._initialize()
        self.options.update(kwargs)

    def _initialize(self) -> None:
        """
        Implemented by sampling methods to declare options (optional).

        Examples
        --------
        self.options.declare('option_name', default_value, types=(bool, int), desc='description')
        """
        pass

    def __call__(self, nt: int) -> np.ndarray:
        """
        Compute the requested number of sampling points.

        The number of dimensions (nx) is determined based on `xlimits.shape[0]`.

        Arguments
        ---------
        nt : int
            Number of points requested.

        Returns
        -------
        ndarray[nt, nx]
            The sampling locations in the input space.
        """
        return self._compute(nt)

    @abstractmethod
    def _compute(self, nt: int) -> np.ndarray:
        """
        Implemented by sampling methods to compute the requested number of sampling points.

        The number of dimensions (nx) is determined based on `xlimits.shape[0]`.

        Arguments
        ---------
        nt : int
            Number of points requested.

        Returns
        -------
        ndarray[nt, nx]
            The sampling locations in the input space.
        """
        raise Exception("This sampling method has not been implemented correctly")


#######################################


def check_xspec_consistency(xtypes, xlimits):
    if len(xlimits) != len(xtypes):
        raise ValueError(
            "number of x limits ({}) do not"
            "correspond to number of specified types ({})".format(len(xlimits), len(xtypes))
        )

    for i, xtyp in enumerate(xtypes):
        if (not isinstance(xtyp, tuple)) and len(xlimits[i]) != 2:
            if xtyp == ORD and isinstance(xlimits[i][0], str):
                listint = list(map(float, xlimits[i]))
                sortedlistint = sorted(listint)
                if not np.array_equal(sortedlistint, listint):
                    raise ValueError(
                        "Unsorted x limits ({}) for variable type {} (index={})".format(xlimits[i], xtyp, i)
                    )

            else:
                raise ValueError("Bad x limits ({}) for variable type {} (index={})".format(xlimits[i], xtyp, i))
        if xtyp == INT:
            if not isinstance(xlimits[i][0], str):
                xtyp = ORD
                xtypes[i] = ORD
            else:
                raise ValueError("INT do not work with list of ordered values, use ORD instead")
        if xtyp != FLOAT and xtyp != ORD and (not isinstance(xtyp, tuple) or xtyp[0] != ENUM):
            raise ValueError("Bad type specification {}".format(xtyp))

        if isinstance(xtyp, tuple) and len(xlimits[i]) != xtyp[1]:
            raise ValueError(
                "Bad x limits and x types specs not consistent. "
                "Got a categorical type with {} levels "
                "while x limits contains {} values (index={})".format(xtyp[1], len(xlimits[i]), i)
            )


def _raise_value_error(xtyp):
    raise ValueError("Bad xtype specification: " "should be FLOAT, ORD or (ENUM, n), got {}".format(xtyp))


def compute_unfolded_dimension(xtypes):
    """
    Returns x dimension (int) taking into account  unfolded categorical features
    """
    res = 0
    for xtyp in xtypes:
        if xtyp == FLOAT or xtyp == ORD:
            res += 1
        elif isinstance(xtyp, tuple) and xtyp[0] == ENUM:
            res += xtyp[1]
        else:
            _raise_value_error(xtyp)
    return res


def unfold_xlimits_with_continuous_limits(xtypes, xlimits, categorical_kernel=None):
    """
    Expand xlimits to add continuous dimensions for enumerate x features
    Each level of an enumerate gives a new continuous dimension in [0, 1].
    Each integer dimensions are relaxed continuously.

    Parameters
    ----------
    xtypes: x types list
        x type specification
    xlimits : list
        bounds of each original dimension (bounds of enumerates being the list of levels).

    Returns
    -------
    np.ndarray [nx continuous, 2]
        bounds of the each dimension where limits for enumerates (ENUM)
        are expanded ([0, 1] for each level).
    """
    # Continuous optimization : do nothing
    xlims = []
    for i, xtyp in enumerate(xtypes):
        if xtyp == FLOAT or xtyp == ORD:
            k = xlimits[i][0]
            if xtyp == ORD and (not isinstance(xlimits[i][0], int)):
                listint = list(map(float, xlimits[i]))
                listint = [listint[0], listint[-1]]
                xlims.append(listint)
            else:
                xlims.append(xlimits[i])
        elif isinstance(xtyp, tuple) and xtyp[0] == ENUM:
            if xtyp[1] == len(xlimits[i]):
                if categorical_kernel is None:
                    xlims.extend(xtyp[1] * [[0, 1]])
                else:
                    listint = list(map(float, [0, len(xlimits[i])]))
                    listint = [listint[0], listint[-1]]
                    xlims.append(listint)
            else:
                raise ValueError(
                    "Bad xlimits for categorical var[{}] "
                    "should have {} categories, got only {} in {}".format(i, xtyp[1], len(xlimits[i]), xlimits[i])
                )
        else:
            _raise_value_error(xtyp)
    return np.array(xlims).astype(float)


def cast_to_discrete_values(xtypes, xlimits, categorical_kernel, x):
    """
    see MixedIntegerContext.cast_to_discrete_values
    """
    ret = ensure_2d_array(x, "x").copy()
    x_col = 0
    for i, xtyp in enumerate(xtypes):
        if xtyp == FLOAT:
            x_col += 1
            continue
        elif xtyp == ORD:
            if isinstance(xlimits[i][0], str):
                listint = list(map(float, xlimits[i]))
                ret[:, x_col] = take_closest_in_list(listint, ret[:, x_col])
            else:
                ret[:, x_col] = np.round(ret[:, x_col])
            x_col += 1
        elif isinstance(xtyp, tuple) and xtyp[0] == ENUM:
            if categorical_kernel is None:
                # Categorial : The biggest level is selected.
                xenum = ret[:, x_col : x_col + xtyp[1]]
                maxx = np.max(xenum, axis=1).reshape((-1, 1))
                mask = xenum < maxx
                xenum[mask] = 0
                xenum[~mask] = 1
                x_col = x_col + xtyp[1]
            else:
                ret[:, x_col] = np.round(ret[:, x_col])
                x_col += 1
        else:
            _raise_value_error(xtyp)
    return ret


def fold_with_enum_index(xtypes, x, categorical_kernel=None):
    """
    see MixedIntegerContext.fold_with_enum_index
    """
    x = np.atleast_2d(x)
    xfold = np.zeros((x.shape[0], len(xtypes)))
    unfold_index = 0
    for i, xtyp in enumerate(xtypes):
        if xtyp == FLOAT or xtyp == ORD:
            xfold[:, i] = x[:, unfold_index]
            unfold_index += 1
        elif isinstance(xtyp, tuple) and xtyp[0] == ENUM:
            index = np.argmax(x[:, unfold_index : unfold_index + xtyp[1]], axis=1)
            xfold[:, i] = index
            unfold_index += xtyp[1]
        else:
            _raise_value_error(xtyp)
    return xfold


def unfold_with_enum_mask(xtypes, x):
    """
    see MixedIntegerContext.unfold_with_enum_mask
    """
    x = np.atleast_2d(x)
    xunfold = np.zeros((x.shape[0], compute_unfolded_dimension(xtypes)))
    unfold_index = 0
    for i, xtyp in enumerate(xtypes):
        if xtyp == FLOAT or xtyp == ORD:
            xunfold[:, unfold_index] = x[:, i]
            unfold_index += 1
        elif isinstance(xtyp, tuple) and xtyp[0] == ENUM:
            enum_slice = xunfold[:, unfold_index : unfold_index + xtyp[1]]
            for row in range(x.shape[0]):
                enum_slice[row, x[row, i].astype(int)] = 1
            unfold_index += xtyp[1]
        else:
            _raise_value_error(xtyp)
    return xunfold


def cast_to_enum_value(xlimits, x_col, enum_indexes):
    """
    see MixedIntegerContext.cast_to_enum_value
    """
    return [xlimits[x_col][index] for index in enum_indexes]


def cast_to_mixed_integer(xtypes, xlimits, x):
    """
    see MixedIntegerContext.cast_to_mixed_integer
    """
    res = []
    for i, xtyp in enumerate(xtypes):
        xi = x[i]
        if xtyp == FLOAT:
            res.append(xi)
        elif xtyp == ORD:
            res.append(int(xi))
        elif isinstance(xtyp, tuple) and xtyp[0] == ENUM:
            res.append(xlimits[i][int(xi)])
        else:
            _raise_value_error(xtyp)
    return res


class MixedIntegerSamplingMethod(SamplingMethod):
    """
    Sampling method decorator that takes an SMT continuous sampling method and
    cast values according x types specification to implement a sampling method
    handling integer (ORD) or categorical (ENUM) features
    """

    def __init__(self, xtypes, xlimits, sampling_method_class, **kwargs):
        """
        Parameters
        ----------
        xtypes: x types list
            x types specification
        xlimits: array-like
            bounds of x features
        sampling_method_class: class name
            SMT sampling method class
        kwargs: options of the given sampling method
            options used to instanciate the SMT sampling method
            with the additional 'output_in_folded_space' boolean option
            specifying if doe output should be in folded space (enum indexes)
            or not (enum masks)
        """
        super()
        check_xspec_consistency(xtypes, xlimits)
        self._xtypes = xtypes
        self._xlimits = xlimits
        self._unfolded_xlimits = unfold_xlimits_with_continuous_limits(self._xtypes, xlimits)
        self._output_in_folded_space = kwargs.get("output_in_folded_space", True)
        kwargs.pop("output_in_folded_space", None)
        self._sampling_method = sampling_method_class(xlimits=self._unfolded_xlimits, **kwargs)

    def _compute(self, nt):
        doe = self._sampling_method(nt)
        unfold_xdoe = cast_to_discrete_values(self._xtypes, self._xlimits, None, doe)
        if self._output_in_folded_space:
            return fold_with_enum_index(self._xtypes, unfold_xdoe)
        else:
            return unfold_xdoe

    def __call__(self, nt):
        return self._compute(nt)


#######################################


class ScaledSamplingMethod(SamplingMethod):
    """This class describes an sample method which generates samples in the unit hypercube.

    The __call__ method does scale the generated samples accordingly to the defined xlimits.
    """

    def __call__(self, nt: int) -> np.ndarray:
        """
        Compute the requested number of sampling points.

        The number of dimensions (nx) is determined based on `xlimits.shape[0]`.

        Arguments
        ---------
        nt : int
            Number of points requested.

        Returns
        -------
        ndarray[nt, nx]
            The sampling locations in the input space.
        """
        return _scale_to_xlimits(self._compute(nt), self.options["xlimits"])

    @abstractmethod
    def _compute(self, nt: int) -> np.ndarray:
        """
        Implemented by sampling methods to compute the requested number of sampling points.

        The number of dimensions (nx) is determined based on `xlimits.shape[0]`.

        Arguments
        ---------
        nt : int
            Number of points requested.

        Returns
        -------
        ndarray[nt, nx]
            The sampling locations in the unit hypercube.
        """
        raise Exception("This sampling method has not been implemented correctly")


def _scale_to_xlimits(samples: np.ndarray, xlimits: np.ndarray) -> np.ndarray:
    """Scales the samples from the unit hypercube to the specified limits.

    Parameters
    ----------
    samples : np.ndarray
        The samples with coordinates in [0,1]
    xlimits : np.ndarray
        The xlimits

    Returns
    -------
    np.ndarray
        The scaled samples.
    """
    nx = xlimits.shape[0]
    for kx in range(nx):
        samples[:, kx] = xlimits[kx, 0] + samples[:, kx] * (xlimits[kx, 1] - xlimits[kx, 0])
    return samples


#######################################


class Random(ScaledSamplingMethod):
    def _compute(self, nt):
        """
        Implemented by sampling methods to compute the requested number of sampling points.

        The number of dimensions (nx) is determined based on `xlimits.shape[0]`.

        Arguments
        ---------
        nt : int
            Number of points requested.

        Returns
        -------
        ndarray[nt, nx]
            The sampling locations in the unit hypercube.
        """
        xlimits = self.options["xlimits"]
        nx = xlimits.shape[0]
        return np.random.rand(nt, nx)


#######################################


class LHS(ScaledSamplingMethod):
    def _initialize(self):
        self.options.declare(
            "criterion",
            "c",
            values=[
                "center",
                "maximin",
                "centermaximin",
                "correlation",
                "c",
                "m",
                "cm",
                "corr",
                "ese",
            ],
            types=str,
            desc="criterion used to construct the LHS design "
            + "c, m, cm and corr are abbreviation of center, maximin, centermaximin and correlation, respectively",
        )
        self.options.declare(
            "random_state",
            types=(type(None), int, np.random.RandomState),
            desc="Numpy RandomState object or seed number which controls random draws",
        )

    def _compute(self, nt):
        """
        Implemented by sampling methods to compute the requested number of sampling points.

        The number of dimensions (nx) is determined based on `xlimits.shape[0]`.

        Arguments
        ---------
        nt : int
            Number of points requested.

        Returns
        -------
        ndarray[nt, nx]
            The sampling locations in the unit hypercube.
        """
        xlimits = self.options["xlimits"]
        nx = xlimits.shape[0]

        if isinstance(self.options["random_state"], np.random.RandomState):
            self.random_state = self.options["random_state"]
        elif isinstance(self.options["random_state"], int):
            self.random_state = np.random.RandomState(self.options["random_state"])
        else:
            self.random_state = np.random.RandomState()

        if self.options["criterion"] != "ese":
            return lhs(
                nx,
                samples=nt,
                criterion=self.options["criterion"],
                random_state=self.random_state,
            )
        elif self.options["criterion"] == "ese":
            return self._ese(nx, nt)

    def _maximinESE(
        self,
        X,
        T0=None,
        outer_loop=None,
        inner_loop=None,
        J=20,
        tol=1e-3,
        p=10,
        return_hist=False,
        fixed_index=[],
    ):
        """

        Returns an optimized design starting from design X. For more information,
        see R. Jin, W. Chen and A. Sudjianto (2005):
        An efficient algorithm for constructing optimal design of computer
        experiments. Journal of Statistical Planning and Inference, 134:268-287.


        Parameters
        ----------

        X : array
            The design to be optimized

        T0 : double, optional
        Initial temperature of the algorithm.
        If set to None, a standard temperature is used.

        outer_loop : integer, optional
        The number of iterations of the outer loop. If None, set to
        min(1.5*dimension of LHS, 30)

        inner_loop : integer, optional
        The number of iterations of the inner loop. If None, set to
        min(20*dimension of LHS, 100)

        J : integer, optional
        Number of replications of the plan in the inner loop. Default to 20

        tol : double, optional
        Tolerance for modification of Temperature T. Default to 0.001

        p : integer, optional
        Power used in the calculation of the PhiP criterion. Default to 10

        return_hist : boolean, optional
        If set to True, the function returns information about the behaviour of
        temperature, PhiP criterion and probability of acceptance during the
        process of optimization. Default to False


        Returns
        ------

        X_best : array
        The optimized design

        hist : dictionnary
        If return_hist is set to True, returns a dictionnary containing the phiP
        ('PhiP') criterion, the temperature ('T') and the probability of
        acceptance ('proba') during the optimization.

        """

        # Initialize parameters if not defined
        if T0 is None:
            T0 = 0.005 * self._PhiP(X, p=p)
        if inner_loop is None:
            inner_loop = min(20 * X.shape[1], 100)
        if outer_loop is None:
            outer_loop = min(int(1.5 * X.shape[1]), 30)

        T = T0
        X_ = X[:]  # copy of initial plan
        X_best = X_[:]
        d = X.shape[1]
        PhiP_ = self._PhiP(X_best, p=p)
        PhiP_best = PhiP_

        hist_T = list()
        hist_proba = list()
        hist_PhiP = list()
        hist_PhiP.append(PhiP_best)

        # Outer loop
        for z in range(outer_loop):
            PhiP_oldbest = PhiP_best
            n_acpt = 0
            n_imp = 0

            # Inner loop
            for i in range(inner_loop):

                modulo = (i + 1) % d
                l_X = list()
                l_PhiP = list()

                # Build J different plans with a single exchange procedure
                # See description of PhiP_exchange procedure
                for j in range(J):
                    l_X.append(X_.copy())
                    l_PhiP.append(self._PhiP_exchange(l_X[j], k=modulo, PhiP_=PhiP_, p=p, fixed_index=fixed_index))

                l_PhiP = np.asarray(l_PhiP)
                k = np.argmin(l_PhiP)
                PhiP_try = l_PhiP[k]

                # Threshold of acceptance
                if PhiP_try - PhiP_ <= T * self.random_state.rand(1)[0]:
                    PhiP_ = PhiP_try
                    n_acpt = n_acpt + 1
                    X_ = l_X[k]

                    # Best plan retained
                    if PhiP_ < PhiP_best:
                        X_best = X_
                        PhiP_best = PhiP_
                        n_imp = n_imp + 1

                hist_PhiP.append(PhiP_best)

            p_accpt = float(n_acpt) / inner_loop  # probability of acceptance
            p_imp = float(n_imp) / inner_loop  # probability of improvement

            hist_T.extend(inner_loop * [T])
            hist_proba.extend(inner_loop * [p_accpt])

            if PhiP_best - PhiP_oldbest < tol:
                # flag_imp = 1
                if p_accpt >= 0.1 and p_imp < p_accpt:
                    T = 0.8 * T
                elif p_accpt >= 0.1 and p_imp == p_accpt:
                    pass
                else:
                    T = T / 0.8
            else:
                # flag_imp = 0
                if p_accpt <= 0.1:
                    T = T / 0.7
                else:
                    T = 0.9 * T

        hist = {"PhiP": hist_PhiP, "T": hist_T, "proba": hist_proba}

        if return_hist:
            return X_best, hist
        else:
            return X_best

    def _PhiP(self, X, p=10):
        """
        Calculates the PhiP criterion of the design X with power p.

        X : array_like
        The design where to calculate PhiP
        p : integer
        The power used for the calculation of PhiP (default to 10)
        """

        return ((pdist(X) ** (-p)).sum()) ** (1.0 / p)

    def _PhiP_exchange(self, X, k, PhiP_, p, fixed_index):
        """
        Modifies X with a single exchange algorithm and calculates the corresponding
        PhiP criterion. Internal use.
        Optimized calculation of the PhiP criterion. For more information, see:
        R. Jin, W. Chen and A. Sudjianto (2005):
        An efficient algorithm for constructing optimal design of computer
        experiments. Journal of Statistical Planning and Inference, 134:268-287.

        Parameters
        ----------

        X : array_like
        The initial design (will be modified during procedure)

        k : integer
        The column where the exchange is proceeded

        PhiP_ : double
        The PhiP criterion of the initial design X

        p : integer
        The power used for the calculation of PhiP


        Returns
        ------

        res : double
        The PhiP criterion of the modified design X

        """

        # Choose two (different) random rows to perform the exchange
        i1 = self.random_state.randint(X.shape[0])
        while i1 in fixed_index:
            i1 = self.random_state.randint(X.shape[0])

        i2 = self.random_state.randint(X.shape[0])
        while i2 == i1 or i2 in fixed_index:
            i2 = self.random_state.randint(X.shape[0])

        X_ = np.delete(X, [i1, i2], axis=0)

        dist1 = cdist([X[i1, :]], X_)
        dist2 = cdist([X[i2, :]], X_)
        d1 = np.sqrt(dist1**2 + (X[i2, k] - X_[:, k]) ** 2 - (X[i1, k] - X_[:, k]) ** 2)
        d2 = np.sqrt(dist2**2 - (X[i2, k] - X_[:, k]) ** 2 + (X[i1, k] - X_[:, k]) ** 2)

        res = (PhiP_**p + (d1 ** (-p) - dist1 ** (-p) + d2 ** (-p) - dist2 ** (-p)).sum()) ** (1.0 / p)
        X[i1, k], X[i2, k] = X[i2, k], X[i1, k]

        return res

    def _ese(self, dim, nt, fixed_index=[], P0=[]):
        """
        Parameters
        ----------

        fixed_index : list
            When running an "ese" optimization, we can fix the indexes of
            the points that we do not want to modify

        """
        # Parameters of maximinESE procedure
        if len(fixed_index) == 0:
            P0 = lhs(dim, nt, criterion=None, random_state=self.random_state)
        else:
            P0 = P0
            self.random_state = np.random.RandomState()
        J = 20
        outer_loop = min(int(1.5 * dim), 30)
        inner_loop = min(20 * dim, 100)

        D0 = pdist(P0)
        R0 = np.corrcoef(P0)
        corr0 = np.max(np.abs(R0[R0 != 1]))
        phip0 = self._PhiP(P0)

        P, historic = self._maximinESE(
            P0,
            outer_loop=outer_loop,
            inner_loop=inner_loop,
            J=J,
            tol=1e-3,
            p=10,
            return_hist=True,
            fixed_index=fixed_index,
        )
        return P

    def expand_lhs(self, x, n_points, method="basic"):
        """
        Given a Latin Hypercube Sample (LHS) "x", returns an expanded LHS
        by adding "n_points" new points.

        Parameters
        ----------
        x : array
            Initial LHS.
        n_points : integer
            Number of points that are to be added to the expanded LHS.
        method : str, optional
            Methodoly for the construction of the expanded LHS.
            The default is "basic". The other option is "ese" to use the
            ese optimization

        Returns
        -------
        x_new : array
            Expanded LHS.

        """

        xlimits = self.options["xlimits"]

        new_num = len(x) + n_points
        if new_num % len(x) != 0:
            print(
                "WARNING: The added number of points is not a "
                "multiple of the initial number of points."
                "Thus, it cannot be ensured that the output is an LHS."
            )

        # Evenly spaced intervals with the final dimension of the LHS
        intervals = []
        for i in range(len(xlimits)):
            intervals.append(np.linspace(xlimits[i][0], xlimits[i][1], new_num + 1))

        # Creates a subspace with the rows and columns that have no points
        # in the new space
        subspace_limits = [[]] * len(xlimits)
        subspace_bool = []
        for i in range(len(xlimits)):
            subspace_limits[i] = []

            subspace_bool.append(
                [
                    [intervals[i][j] < x[kk][i] < intervals[i][j + 1] for kk in range(len(x))]
                    for j in range(len(intervals[i]) - 1)
                ]
            )

            [
                subspace_limits[i].append([intervals[i][ii], intervals[i][ii + 1]])
                for ii in range(len(subspace_bool[i]))
                if not (True in subspace_bool[i][ii])
            ]

        # Sampling of the new subspace
        sampling_new = LHS(xlimits=np.array([[0.0, 1.0]] * len(xlimits)))
        x_subspace = sampling_new(n_points)

        column_index = 0
        sorted_arr = x_subspace[x_subspace[:, column_index].argsort()]

        for j in range(len(xlimits)):
            for i in range(len(sorted_arr)):
                sorted_arr[i, j] = subspace_limits[j][i][0] + sorted_arr[i, j] * (
                    subspace_limits[j][i][1] - subspace_limits[j][i][0]
                )

        H = np.zeros_like(sorted_arr)
        for j in range(len(xlimits)):
            order = np.random.permutation(len(sorted_arr))
            H[:, j] = sorted_arr[order, j]

        x_new = np.concatenate((x, H), axis=0)

        if method == "ese":
            # Sampling of the new subspace
            sampling_new = LHS(xlimits=xlimits, criterion="ese")
            x_new = sampling_new._ese(len(x_new), len(x_new), fixed_index=np.arange(0, len(x), 1), P0=x_new)

        return x_new


#######################################



from typing import Dict, Union

import numpy as np
import pandas as pd
from sklearn.metrics import r2_score


def paritydata(Y_actual: np.ndarray, Y_predicted: np.ndarray) -> pd.DataFrame:
    """

    Creates a pd.DataFrame that contains the actual and predicted values

    """
    if Y_actual.ndim == 1:
        Y_actual = Y_actual.reshape(-1, 1)

    if len(Y_predicted.shape) == 1:
        Y_predicted = Y_predicted.reshape(-1, 1)

    return pd.DataFrame(np.hstack((Y_actual, Y_predicted)), columns=["Actual", "Predicted"])


def rmse(Y_actual: np.ndarray, Y_predicted: np.ndarray) -> np.ndarray:
    """

    Determines the root mean square error (RMSE) between the actual and predicted values

    """
    if Y_actual.ndim > 1:
        Y_actual = Y_actual.flatten()

    if len(Y_predicted.shape) > 1:
        Y_predicted = Y_predicted.flatten()

    return np.sqrt(np.mean((Y_actual - Y_predicted) ** 2))


def r2(Y_actual: np.ndarray, Y_predicted: np.ndarray) -> np.ndarray:
    """

    Determines the coefficient of determination (r^2) between the actual and predicted values

    """
    if Y_actual.ndim > 1:
        Y_actual = Y_actual.flatten()

    if len(Y_predicted.shape) > 1:
        Y_predicted = Y_predicted.flatten()

    return r2_score(Y_actual, Y_predicted)


def all_performance_metrics(Y: np.ndarray, Y_predicted: np.ndarray) -> Dict[str, Union[np.ndarray, pd.DataFrame]]:
    """

    Creates a dictionary containing the root mean square error (RMSE), coefficient of determination (r^2), and
    pd.DataFrame of the actual and predicted values

    """
    return {
        "RMSE": rmse(Y, Y_predicted),
        "r2": r2(Y, Y_predicted),
        "Parity Data": paritydata(Y, Y_predicted),
    }


def all_performance_metrics_train_val(
    Y_train: np.ndarray, Y_train_predicted: np.ndarray, Y_val: np.ndarray, Y_val_predicted: np.ndarray
) -> Dict[str, Union[np.ndarray, pd.DataFrame]]:
    """

    Creates a dictionary containing the root mean square error (RMSE), coefficient of determination (r^2), and
    pd.DataFrame of the actual and predicted values for training and validation data

    Parameters
    ----------
    Y_train: np.ndarray
        Actual values for training data
    Y_train_predicted: np.ndarray
        Predicted values for training data
    Y_val: np.ndarray
        Actual values for validation data
    Y_val_predicted: np.ndarray
        Predicted values for validation data

    """
    return {
        "Train RMSE": rmse(Y_train, Y_train_predicted),
        "Validation RMSE": rmse(Y_val, Y_val_predicted),
        "Train r2": r2(Y_train, Y_train_predicted),
        "Validation r2": r2(Y_val, Y_val_predicted),
        "Train Parity Data": paritydata(Y_train, Y_train_predicted),
        "Validation Parity Data": paritydata(Y_val, Y_val_predicted),
    }


def all_performance_metrics_train_val_test(
    Y_train: np.ndarray,
    Y_train_predicted: np.ndarray,
    Y_val: np.ndarray,
    Y_val_predicted: np.ndarray,
    Y_test: np.ndarray,
    Y_test_predicted: np.ndarray,
) -> Dict[str, Union[np.ndarray, pd.DataFrame]]:
    """

    Creates a dictionary containing the root mean square error (RMSE), coefficient of determination (r^2), and
    pd.DataFrame of the actual and predicted values for training, validation, and test data

    Parameters
    ----------
    Y_train: np.ndarray
        Actual values for training data
    Y_train_predicted: np.ndarray
        Predicted values for training data
    Y_val: np.ndarray
        Actual values for validation data
    Y_val_predicted: np.ndarray
        Predicted values for validation data
    Y_test: np.ndarray
        Actual values for test data
    Y_test_predicted: np.ndarray
        Predicted values for test data

    """
    return {
        "Train RMSE": rmse(Y_train, Y_train_predicted),
        "Validation RMSE": rmse(Y_val, Y_val_predicted),
        "Test RMSE": rmse(Y_test, Y_test_predicted),
        "Train r2": r2(Y_train, Y_train_predicted),
        "Validation r2": r2(Y_val, Y_val_predicted),
        "Test r2": r2(Y_test, Y_test_predicted),
        "Train Parity Data": paritydata(Y_train, Y_train_predicted),
        "Validation Parity Data": paritydata(Y_val, Y_val_predicted),
        "Test Parity Data": paritydata(Y_test, Y_test_predicted),
    }



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nemo_bo',
 'nemo_bo.acquisition_functions',
 'nemo_bo.acquisition_functions.expected_improvement',
 'nemo_bo.acquisition_functions.highest_uncertainty',
 'nemo_bo.acquisition_functions.nsga_improvement',
 'nemo_bo.models',
 'nemo_bo.models.base',
 'nemo_bo.opt',
 'nemo_bo.opt.utils_samplers',
 'nemo_bo.utils']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=21.4.0,<22.0.0',
 'botorch>=0.7.0,<0.8.0',
 'cython>=0.29.30,<0.30.0',
 'forestci>=0.5.1,<0.6.0',
 'gpytorch>=1.8.1,<2.0.0',
 'hyperopt>=0.2.7,<0.3.0',
 'matplotlib>=3.5.2,<4.0.0',
 'ngboost>=0.3.12,<0.4.0',
 'numba>=0.55.1,<0.56.0',
 'numpy>=1.22.4,<2.0.0',
 'openpyxl>=3.0.9,<4.0.0',
 'pandas>=1.4.2,<2.0.0',
 'pydoe2>=1.3.0,<2.0.0',
 'pymoo>=0.6.0,<0.7.0',
 'scikit-learn>=1.1.1,<2.0.0',
 'scipy>=1.8.1,<2.0.0',
 'six>=1.16.0,<2.0.0',
 'tensorflow-probability>=0.17.0,<0.18.0',
 'tensorflow>=2.9.1,<3.0.0',
 'torch>=1.11.0,<2.0.0',
 'xgboost-distribution>=0.2.4,<0.3.0']

setup_kwargs = {
    'name': 'nemo-bo',
    'version': '0.1.13',
    'description': 'Multi-objective optimization of chemical processes with automated machine learning workflows',
    'long_description': '# Nomadic Exploratory Multi-objective Optimisation (NEMO)\n\nMeet NEMO - our ‘Nomadic Explorer’. NEMO is quite the connoisseur when it comes to machine learning optimisation - only the best model types and model parameters will suffice. In the case of a single dataset, NEMO will scour the lands for the optimal model type and parameters to fit a given dataset. A range of outputs will then be generated for you to assess, interpret and utilise your newly created model.\n\nIf you decide to take your analyses a step further into the realms of Multi-objective Bayesian optimisation, then our Nomadic Explorer will tirelessly search for the best model type and parameters at each iteration of the optimisation. At each stage, the optimal set of conditions will be provided to aid your pursuit of the elusive multi-dimensional pareto front.\n\nNEMO is prepared for the journey with a cavernous bag of tools. However, if your aspirations are more exotic, then NEMO supports the inclusion of custom models, samplers and functions.\n\nCheck out the examples to see NEMO in action and get started with your own ML workflows.\n\n## What is NEMO?\n\nNEMO is a package designed for Bayesian optimisation of one or multiple objectives simultaneously, with a focus on applying to chemical processes.\n\n## Installation\nTo install NEMO via pip:\n\n`pip install nemo-bo`\n\n\n## How does NEMO work?\n\nFirstly, the parameters (variables) and targets (objectives) of a chemical process are provided to the algorithm. After \nproviding NEMO with a dataset from prior experiments, it will then identify the relationship between the parameters and targets \nand then suggest the ideal parameters to use for the optimisation iteration.\n\nIn comparison to other open-source optimisation libraries, NEMO will automatically optimise the hyperparameters for \nvarious machine learning models and select the one with the best predictive accuracy for a given objective. This \nensures that the model is continuously optimised over the course of an optimisation campaign. Furthermore, NEMO \nnatively supports objectives that can be calculated if the exact relationship between the parameters and the target \n(e.g. materials cost) is known.\n\n## What features are in NEMO?\nAlthough NEMO includes many machine learning models, acquisition functions, constraints, and sample generators, the \nbase classes for these are all included and can be utilised as a template for adding your own custom solutions.\n\nThe features natively found in NEMO are the following:\n\n### Machine learning models available\n1. Gaussian processes (GPs) using the [BoTorch](https://github.com/pytorch/botorch) library\n\n2. Various neural networks from the [Deeply Uncertain code repository](https://github.com/deepskies/DeeplyUncertain-Public):\n   1. Bayesian neural networks\n   2. Concrete dropout\n   3. Deep ensembles\n\n3. Various decision-tree based models:\n   1. [XGBoost Distribution](https://github.com/CDonnerer/xgboost-distribution)\n   2. [NGBoost](https://github.com/stanfordmlgroup/ngboost)\n   3. Random Forest using the [forest-confidence-interval](https://github.com/scikit-learn-contrib/forest-confidence-interval)\n\n### Variable types available\n1. Continuous variables (`ContinuousVariable`)\n2. Categorical variables with discrete variables (`CategoricalVariableDiscreteValues`)\n3. Categorical variables with descriptors (`CategoricalVariableWithDescriptors`)\n\nCategorical variables without any description (e.g. one-hot encoding) is not currently supported\n\n### Objective types available\n1. Objectives modelled using machine learning models (`RegressionObjective`)\n2. Calculated objectives using a user-provided function (`CalculableObjective`)\n\nClassification objectives are not currently supported\n\n### User-selectable acquisition functions available\n1. Expected improvement based methods (`ExpectedImprovement`)\n   1. A modifed single-objective expected improvement algorithm that is better at exploration than the standard analytical method\n   2. A modifed multi-objective expected hypervolume improvement algorithm that is better at exploration than the standard analytical method\n   3. [qNEI](https://botorch.org/api/_modules/botorch/acquisition/monte_carlo.html#qNoisyExpectedImprovement) and [qNEHVI](https://botorch.org/api/_modules/botorch/acquisition/multi_objective/monte_carlo.html#qNoisyExpectedHypervolumeImprovement) BoTorch methods (only compatible with GP models)\n\n2. A Unified evolutionary optimization algorithm [U-NSGA-III](https://pymoo.org/algorithms/moo/unsga3.html) based method that derives uncertainty in the inference by sampling from a distribution (`NSGAImprovement`)\n3. A fully explorative method that identifies the candidates that have the highest uncertainty in the objective predictions (`HighestUncertainty`)\n\n### Input constraints available\n1. Linear equality and inequality constraints(`LinearConstraint`)\n2. Basic non-linear equality and inequality constraints that incorporates an exponent for each input variable (`NonLinearPowerConstraint`)\n3. Equality and inequality constraints that allows the user to pass a function to calculate the left-hand-side of the constraint (`FunctionalConstraint`)\n4. Stoichiometry constraints that forces the ratio between two input variable to be equal to or greater than a specified value (`StoichiometricConstraint`)\n5. A constraint type to limit the number of active variables (`MaxActiveFeaturesConstraint`)\n6. A constraint type that prevents certain categorical constraints from being selected simulatenously (`CategoricalConstraint`)\n\n### Benchmarking functionality available\nBenchmark functions are typically used to simulate the outcomes of experiments in a closed-loop manner and therefore\ncan be helpful to evaluate the quality of an optimisation (inferred from the effectiveness of the utilised model(s) \nand/or acquisition function to identify the optimum)\n\n1. Machine learning model based on a provided dataset (`ModelBenchmark`)\n2. Single objective synthetic functions (`SingleObjectiveSyntheticBenchmark`)\n3. Multi-objective synthetic functions (`MultiObjectiveSyntheticBenchmark`)\n\n### Sample generators available\nMethods for generating a samples of parameter values during an optimisation. These can be used independently outside of an optimisation too by calling the `generate_samples` function\n\n1. Latin hypercube sampling (with a mixed-integer implementation for efficient sampling of categorical variables) (`LatinHyperCubeSampling`)\n2. Sobol sampling (`SobolSampling`)\n3. Polytope sampling (`PolytopeSampling`)\n4. Random sampling (`RandomSampling`)\n5. Pool-based sampling using a user-defined set of data points. Typically used as an alternative to a machine learning model benchmark function (`PoolBased`)\n\n### Other utilities/functions available\n1. Included template for provided the dataset with automated extraction\n2. Scatter and bar chart plotting functionality for displaying model quality and optimisation progress\n\n\n## Getting started\nThe following code demonstrates how to set-up a simple bayesian optimisation using a user-provided dataset containing two continuous variables (X) and two objectives (Y):\n\n```python\n# Import the variable, objectives, sampler, acquisition function, and the optimisation classes\nfrom nemo_bo.opt.variables import ContinuousVariable, VariablesList\nfrom nemo_bo.opt.objectives import RegressionObjective, ObjectivesList\nfrom nemo_bo.acquisition_functions.expected_improvement.expected_improvement import ExpectedImprovement\nfrom nemo_bo.opt.samplers import LatinHyperCubeSampling\nfrom nemo_bo.opt.optimisation import Optimisation\n\n# Create the variable objects\nvar1 = ContinuousVariable(name="variable1", lower_bound=0.0, upper_bound=100.0)\nvar2 = ContinuousVariable(name="variable2", lower_bound=0.0, upper_bound=100.0)\nvar_list = VariablesList([var1, var2])\n\n# Create the objective objects\nobj1 = RegressionObjective(name="objective1", obj_max_bool=True, lower_bound=0.0, upper_bound=100.0) # obj_max_bool when True defines the objective is to be maximised\nobj2 = RegressionObjective(name="objective2", obj_max_bool=False, lower_bound=0.0, upper_bound=100.0) # obj_max_bool when False defines the objective is to be minimised\nobj_list = ObjectivesList([obj1, obj2])\n\n# Instantiate the sampler\nsampler = LatinHyperCubeSampling()\n\n# Instantiate the acquisition function\nacq_func = ExpectedImprovement(num_candidates=4) # num_candidates defines how many sets of parameters to return at each optimisation iteration\n\n# Set up the optimisation instance\noptimisation = Optimisation(var_list, obj_list, acq_func, sampler=sampler)\n\n# Start the optimisation using the convenient run function that will run for the specified number of iterations\n# X and Y arrays represent a hypothetical user-provided dataset\noptimisation_data = optimisation.run(X, Y, number_of_iterations=50)\n```\n\n## More tutorials\nWe encourage you to look through the tutorials written in the `tutorials` folder to see how to use some other \nNEMO functions\n1. How to select specific machine learning models types for the objectives\n2. Setting up a single objective optimisation\n3. How to use calculable objectives\n4. How to define transformers for variables and objectives\n5. How to define categorical variables with descriptors\n6. Utilising the machine learning model fitting in NEMO without Bayesian optimisation\n7. How to create a closed-loop optimisation using a machine learning model as the benchmark function\n8. How to create a closed-loop optimisation using a multiobjective synthetic function as the benchmark function\n9. How to create a closed-loop optimisation using a single objective synthetic function as the benchmark function\n10. How to create a closed-loop optimisation using a pool-based sampler as the benchmark\n11. Setting up an optimisation with input constraints\n12. Generating samples without needing to perform an optimisation\n13. How to set up a manual optimisation\n14. How to resume an optimisation run\n15. How to use the BoTorch (quasi-) Monte-Carlo based acquisition functions in NEMO\n16. How to set up an optimisation that uses U-NSGA-III as the acquisition function\n17. Using the input template excel file template to import the variables and objectives data\n18. How to set up an optimisation that uses the highest uncertainty acquisition function\n\n\n## What to do if you find any issues?\nLeave a message in the issues section and we will get back to you as soon as we can.\n\n\n## Acknowledgements\nMuch of the functionality in NEMO is built on top of the work by the authors of the features we incorporate. We are grateful \nto them for continuously supporting their libraries and establishing their platforms for optimisation work. We reference \nthe works throughout the .py files.\n',
    'author': 'Simon Sung',
    'author_email': 'simon.sung06@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/sustainable-processes/nemo-bo',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9.0,<=3.9.13',
}


setup(**setup_kwargs)



import os
from typing import Any, List, Optional, Tuple, Union

import matplotlib.animation as animation
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib.animation import FuncAnimation
from matplotlib.axes import Axes


def scatter_plot(
    ax: Axes,
    data: Union[List[np.ndarray], np.ndarray],
    legend: Union[str, List[str]],
    error: Optional[Union[List[np.ndarray], np.ndarray]] = None,
) -> None:
    """

    Function to create the scatter plot Axes object

    Parameters
    ----------
    ax: Axes
        Matplotlib Axes object
    data: List[np.ndarray] | np.ndarray
        Can be a 2D (x, y) or 3D (x, y, z) array or a list of 2D or 3D arrays for drawing multiple plots on to the same
        set of axes
    legend: str | List[str]
        A string or a list of strings for labelling the plots
    error: List[np.ndarray] | np.ndarray, Default = None
        Can be an array or a list of arrays that contains the respective values to be used for error bars

    """
    # If the data provided is an array (i.e. not a list of arrays)
    # Starts the colour cycle for the plots from a different position (brown) to the default (blue)
    colorcycle = [
        "#1f77b4",
        "#ff7f0e",
        "#2ca02c",
        "#d62728",
        "#9467bd",
        "#8c564b",
        "#e377c2",
        "#7f7f7f",
        "#bcbd22",
        "#17becf",
    ]
    if np.array(data)[0][0].ndim == 0:
        data_plot = np.array(data)
        # If the array provided is 2D
        if data_plot.shape[1] == 2:
            # If legend information was provided
            if error is None:
                if legend is not None:
                    ax.scatter(
                        data_plot[:, 0],
                        data_plot[:, 1],
                        label=legend,
                        marker="o",
                        s=40,
                        facecolors="none",
                        edgecolors=colorcycle[0],
                    )
                else:
                    ax.scatter(
                        data_plot[:, 0],
                        data_plot[:, 1],
                        label="Scatter plot",
                        marker="o",
                        s=40,
                        facecolors="none",
                        edgecolors=colorcycle[0],
                    )
            else:
                if legend is not None:
                    ax.errorbar(
                        data_plot[:, 0],
                        data_plot[:, 1],
                        yerr=error.flatten(),
                        fmt="o",
                        label=legend,
                        zorder=2,
                        fillstyle="none",
                    )
                else:
                    ax.errorbar(
                        data_plot[:, 0],
                        data_plot[:, 1],
                        yerr=error.flatten(),
                        fmt="o",
                        label="Scatter plot",
                        zorder=2,
                        fillstyle="none",
                    )
        # Else if the array provided is 3D
        elif data_plot.shape[1] == 3:
            # If legend information was provided
            if legend is not None:
                ax.scatter(
                    data_plot[:, 0],
                    data_plot[:, 1],
                    data_plot[:, 2],
                    label=legend,
                    marker="o",
                    s=40,
                )
            else:
                ax.scatter(
                    data_plot[:, 0],
                    data_plot[:, 1],
                    data_plot[:, 2],
                    label="Scatter plot",
                    marker="o",
                    s=40,
                )
    # Else the data provided is a list of arrays
    else:
        # Iterate through the list of arrays and plot each one
        for plotnum in range(len(data)):
            data_plot = np.array(data[plotnum])
            # If the array provided is 2D
            if data_plot.shape[1] == 2:
                # If legend information was provided
                if error is None:
                    if legend is not None:
                        ax.scatter(
                            data_plot[:, 0],
                            data_plot[:, 1],
                            label=legend[plotnum],
                            marker="o",
                            s=40,
                            facecolors="none",
                            edgecolors=colorcycle[plotnum],
                        )
                    else:
                        ax.scatter(
                            data_plot[:, 0],
                            data_plot[:, 1],
                            label=error_plot.flatten(),
                            marker="o",
                            s=40,
                            facecolors="none",
                            edgecolors=colorcycle[plotnum],
                        )
                else:
                    error_plot = np.array(error[plotnum])
                    if legend is not None:
                        ax.errorbar(
                            data_plot[:, 0],
                            data_plot[:, 1],
                            yerr=error_plot.flatten(),
                            fmt="o",
                            label=legend[plotnum],
                            zorder=2,
                            fillstyle="none",
                        )
                    else:
                        ax.errorbar(
                            data_plot[:, 0],
                            data_plot[:, 1],
                            yerr=error_plot.flatten(),
                            fmt="o",
                            label=f"Scatter plot {plotnum + 1}",
                            zorder=2,
                            fillstyle="none",
                        )
            # Else if the array provided is 3D
            elif data_plot.shape[1] == 3:
                # If legend information was provided
                if legend is not None:
                    ax.scatter(
                        data_plot[:, 0],
                        data_plot[:, 1],
                        data_plot[:, 2],
                        label=legend[plotnum],
                        marker="o",
                        s=40,
                    )
                else:
                    ax.scatter(
                        data_plot[:, 0],
                        data_plot[:, 1],
                        data_plot[:, 2],
                        label=f"Scatter plot {plotnum + 1}",
                        marker="o",
                        s=40,
                    )


def line_plot(data: Union[List[np.ndarray], np.ndarray], legend: Union[str, List[str]]) -> None:
    """

    Function to create the line plot object

    Parameters
    ----------
    data: List[np.ndarray] | np.ndarray
        Can be a 2D (x, y) or 3D (x, y, z) array or a list of 2D or 3D arrays for drawing multiple plots on to the same
        set of axes
    legend: str | List[str]
        A string or a list of strings for labelling the plots

    """
    # Starts the colour cycle for the plots from a different position (brown) to the default (blue)
    colorcycle = [
        "#8c564b",
        "#e377c2",
        "#7f7f7f",
        "#bcbd22",
        "#17becf",
        "#1f77b4",
        "#ff7f0e",
        "#2ca02c",
        "#d62728",
        "#9467bd",
    ]
    # If the data provided is a array (i.e. not a list of arrays)
    if np.array(data)[0][0].ndim == 0:
        data_plot = np.array(data)
        # If the array provided is 2D
        if data_plot.shape[1] == 2:
            # If legend information was provided
            if legend is not None:
                plt.plot(data_plot[:, 0], data_plot[:, 1], label=legend, color=colorcycle[0])
            else:
                plt.plot(
                    data_plot[:, 0],
                    data_plot[:, 1],
                    label="Line plot",
                    color=colorcycle[0],
                )
        # Else if the array provided is 3D
        elif data_plot.shape[1] == 3:
            # If legend information was provided
            if legend is not None:
                plt.plot(
                    data_plot[:, 0],
                    data_plot[:, 1],
                    data_plot[:, 2],
                    label=legend,
                    color=colorcycle[0],
                )
            else:
                plt.plot(
                    data_plot[:, 0],
                    data_plot[:, 1],
                    data_plot[:, 2],
                    label="Line plot",
                    color=colorcycle[0],
                )
    # Else the data provided is a list of arrays
    else:
        # Iterate through the list of arrays and plot each one
        for plotnum in range(len(data)):
            data_plot = np.array(data[plotnum])
            # If the array provided is 2D
            if data_plot.shape[1] == 2:
                # If legend information was provided
                if legend is not None:
                    plt.plot(
                        data_plot[:, 0],
                        data_plot[:, 1],
                        label=legend[plotnum],
                        color=colorcycle[plotnum],
                    )
                else:
                    plt.plot(
                        data_plot[:, 0],
                        data_plot[:, 1],
                        label=f"Line plot {plotnum + 1}",
                        color=colorcycle[plotnum],
                    )
            # Else if the array provided is 3D
            elif data_plot.shape[1] == 3:
                # If legend information was provided
                if legend is not None:
                    plt.plot(
                        data_plot[:, 0],
                        data_plot[:, 1],
                        data_plot[:, 2],
                        label=legend[plotnum],
                        color=colorcycle[0],
                    )
                else:
                    plt.plot(
                        data_plot[:, 0],
                        data_plot[:, 1],
                        data_plot[:, 2],
                        label=f"Line plot {plotnum + 1}",
                        color=colorcycle[0],
                    )


def surface_plot(ax: Axes, data: Union[List[np.ndarray], np.ndarray], legend: Union[str, List[str]]) -> None:
    """

    Function to create the scatter plot Axes object

    Parameters
    ----------
    ax: Axes
        Matplotlib Axes object
    data: List[np.ndarray] | np.ndarray
        Can be a 3D (x, y, z) array or a list of 3D arrays for drawing multiple plots on to the same set of axes
    legend: str | List[str]
        A string or a list of strings for labelling the plots

    """
    # Starts the colour cycle for the plots from a different position (brown) to the default (blue)
    colorcycle = [
        "#8c564b",
        "#e377c2",
        "#7f7f7f",
        "#bcbd22",
        "#17becf",
        "#1f77b4",
        "#ff7f0e",
        "#2ca02c",
        "#d62728",
        "#9467bd",
    ]
    # If the data provided is a array (i.e. not a list of arrays)
    if np.array(data)[0][0].ndim == 0:
        data_plot = np.array(data)
        # plot_trisurf breaks the plot when there are fewer than 3 points
        if data_plot.shape[0] >= 3:
            # If legend information was provided
            if legend is not None:
                surf = ax.plot_trisurf(
                    data_plot[:, 0],
                    data_plot[:, 1],
                    data_plot[:, 2],
                    label=legend,
                    alpha=0.3,
                    color=colorcycle[0],
                    antialiased=False,
                )
                surf._facecolors2d = surf._facecolor3d
                surf._edgecolors2d = surf._edgecolor3d
            else:
                surf = ax.plot_trisurf(
                    data_plot[:, 0],
                    data_plot[:, 1],
                    data_plot[:, 2],
                    label=f"Surface plot",
                    alpha=0.3,
                    color=colorcycle[0],
                    antialiased=False,
                )
                surf._facecolors2d = surf._facecolor3d
                surf._edgecolors2d = surf._edgecolor3d
    # Else the data provided is a list of arrays
    else:
        # Iterate through the list of arrays and plot each one
        for plotnum in range(len(data)):
            data_plot = np.array(data[plotnum])
            # plot_trisurf breaks the plot when there are fewer than 3 points
            if data_plot.shape[0] >= 3:
                # If legend information was provided
                if legend is not None:
                    surf = ax.plot_trisurf(
                        data_plot[:, 0],
                        data_plot[:, 1],
                        data_plot[:, 2],
                        label=legend[plotnum],
                        alpha=0.3,
                        color=colorcycle[plotnum],
                        antialiased=False,
                    )
                    surf._facecolors2d = surf._facecolor3d
                    surf._edgecolors2d = surf._edgecolor3d
                else:
                    surf = ax.plot_trisurf(
                        data_plot[:, 0],
                        data_plot[:, 1],
                        data_plot[:, 2],
                        label=f"Surface plot {plotnum + 1}",
                        alpha=0.3,
                        color=colorcycle[plotnum],
                        antialiased=False,
                    )
                    surf._facecolors2d = surf._facecolor3d
                    surf._edgecolors2d = surf._edgecolor3d


def plot3d_video(ax: Axes, output_file: str) -> None:
    """

    Creates a .mp4 video file of a rotating 3D Matplotlib plot using a series of images

    Parameters
    ----------
    ax: Axes
        Matplotlib Axes object
    output_file: str
        String of the output video location and file name. Do not need the video file extension in this string

    """
    angles = np.linspace(0, 360, 91)[:-1]  # Generates a list of 90 angles between 0 and 360 degrees

    # Creates a series of images at different angles and a list of image file names
    imagefiles_list = []
    print(f"Creating temporary image files")
    for i, angle in enumerate(angles):
        ax.view_init(elev=None, azim=angle)  # elevation set to None
        fname = f"tmpimg_{i}.jpeg"
        ax.figure.savefig(fname, dpi=400)  # dpi set to 400
        imagefiles_list.append(fname)

    # Uses mencoder.exe in the home directory to produce a .mp4 movie from a list of image files.
    command = (
        f'mencoder "mf://{",".join(imagefiles_list)}" -mf fps={10} -o {output_file}.mp4 -ovc lavc -lavcopts '
        f"vcodec=msmpeg4v2:vbitrate={8000}"
    )  # fps=10, bitrate=8000
    os.system(command)

    # Deleted the temporary image files created above
    for f in imagefiles_list:
        os.remove(f)


def plot(
    plot_dim: str = "2D",
    scatter_data: Optional[Union[List[Union[np.ndarray, pd.DataFrame]], np.ndarray, pd.DataFrame]] = None,
    error: Optional[Union[List[np.ndarray], np.ndarray]] = None,
    line_data: Optional[Union[List[Union[np.ndarray, pd.DataFrame]], np.ndarray, pd.DataFrame]] = None,
    surface_data: Optional[Union[List[Union[np.ndarray, pd.DataFrame]], np.ndarray, pd.DataFrame]] = None,
    animation_scatter_data: Optional[Union[np.ndarray, pd.DataFrame]] = None,
    animated_scatter_fps: int = 2,
    scatter_legend: Optional[Union[str, List[str], Tuple[str]]] = None,
    line_legend: Optional[Union[str, List[str]]] = None,
    surface_legend: Optional[Union[str, List[str]]] = None,
    xlabel: str = "x",
    ylabel: str = "y",
    zlabel: str = "z",
    plottitle: str = "Title",
    output_file: str = "Plot",
    write_mp4: bool = False,
) -> None:
    """

    Function to start the plotting of a new 2D or 3D figure (including animated 2D scatter plots)

    Parameters
    ----------
    plot_dim: str, Default = '2D'
        Refers to the number of dimensions for the desired plot
    scatter_data: List[np.ndarray | pd.DataFrame] | np.ndarray | pd.DataFrame, Default is None
        The data used for the scatter plot can be either a 2D (x, y) or 3D (x, y, z) array or a list of 2D or 3D arrays
    line_data: List[np.ndarray | pd.DataFrame] | np.ndarray | pd.DataFrame, Default = None
        The data used for the line plot can be either a 2D (x, y) or 3D (x, y, z) array or a list of 2D or 3D arrays
    surface_data: List[np.ndarray | pd.DataFrame] | np.ndarray | pd.DataFrame, Default is None
        The data used for the surface plot can be either a 3D (x, y, z) array or a list of 3D arrays
    animated_scatter_data: np.ndarray | pd.DataFrame, Default = 2
        The 2D (x, y) data to be animated one point at a time
    animated_scatter_fps: int, Default is None
        The number of frames per second to animate at for the animated scatter data
    scatter_legend: str | List[str] | Tuple[str], Default is None
        The string or list of strings used to label the new scatter plot(s)
    line_legend: str | List[str], Default is None.
        The string or list of strings used to label the new line plot(s)
    surface_legend: str | List[str] | Tuple[str], Default is None
        The string or list of strings used to label the new surface plot(s)
    xlabel: str, Default is "x"
        Label of the x-axis
    ylabel: str, Default is "y"
        Label of the y-axis
    zlabel: str, Default is "z"
        Label of the z-axis
    plottitle: str, Default is "Title"
        The title of the figure
    output_file: str, Default is "Plot"
        The output file location of the .png and/or .mp4 of the new figure. Do not need the file extension
        (.png or .mp4)
    write_mp4: bool, Default is False
        When set to True, creates a .mp4 video file of a rotating 3D Matplotlib plot using a series of in-situ
        generated images

    """
    # Creates the Figure and Axes Matplotlib objects, with settings depending on if a 2D or 3D figure is requested
    if plot_dim == "2D":
        fig = plt.figure(figsize=(8, 8))
        ax = fig.add_subplot(1, 1, 1)
    elif plot_dim == "3D":
        fig = plt.figure()
        ax = fig.add_subplot(111, projection="3d")

    # Creates the scatter, line, or surface plot Axes objects respectively, depending on if the data for each type of
    # plot was provided
    if scatter_data is not None:
        scatter_data = data_to_np(scatter_data)
        scatter_plot(ax, scatter_data, scatter_legend, error)
    if line_data is not None:
        line_data = data_to_np(line_data)
        line_plot(line_data, line_legend)
    if surface_data is not None:
        surface_data = data_to_np(surface_data)
        surface_plot(ax, surface_data, surface_legend)

    # Modifies the style settings for the figure, with settings depending on if a 2D or 3D figure is requested
    if plot_dim == "2D":
        ax.set_xlabel(xlabel, fontsize=12)
        ax.set_ylabel(ylabel, fontsize=12)
        ax.set_title(plottitle, fontsize=12)
        plt.legend(loc="best", fontsize=10)
    elif plot_dim == "3D":
        ax.set_xlabel(xlabel, fontsize=6)
        ax.set_ylabel(ylabel, fontsize=6)
        ax.set_zlabel(zlabel, fontsize=6)
        ax.tick_params(axis="both", which="major", labelsize=6)
        ax.set_title(plottitle, pad=24, fontsize=6)
        plt.legend(loc="best", fontsize=6)

    # Additional style settings for the figure
    ax.set_axisbelow(True)
    plt.grid(b=True, which="major", c="whitesmoke")

    if animation_scatter_data is not None:
        if plot_dim != "2D":
            raise TypeError("Animated scatter plots are currently only supported for 2D scatter plots (plot_dim='2D')")

        animation_scatter_data = data_to_np(animation_scatter_data)
        (graph,) = plt.plot([], [], "o", color="red", markersize=10)

        def animate(i):
            graph.set_data(animation_scatter_data[: i + 1, 0], animation_scatter_data[: i + 1, 1])
            return graph

        ani = FuncAnimation(fig, animate, frames=animation_scatter_data.shape[0], repeat=False, interval=500)
        ani.save(f"{output_file}.mp4", writer=animation.FFMpegWriter(fps=animated_scatter_fps), dpi=400)

    # Writes a .png image file of the figure
    plt.savefig(f"{output_file}.png", dpi=400)

    # Writes a .mp4 video file of a rotating 3D Matplotlib plot
    if plot_dim == "3D":
        if write_mp4 == True:
            plot3d_video(ax, output_file)


def data_to_np(data: Union[List[Union[np.ndarray, pd.DataFrame]], np.ndarray, pd.DataFrame]) -> np.ndarray:
    """

    Function used to convert pd.DataFrames, np.ndarray or a list of pd.DataFrame or np.ndarray into 2D or 3D np.ndarray
    objects

    Parameters
    ----------
    data: List[np.ndarray | pd.DataFrame] | np.ndarray | pd.DataFrame
        Any 2D or 3D array or list of 2D or 3D array

    """
    inputdata = data
    if isinstance(data, list):
        if isinstance(data[0], pd.DataFrame) or isinstance(data[0], np.ndarray):
            inputdata = np.zeros(len(data), dtype=np.dtype("object"))
            for x in range(len(inputdata)):
                inputdata[x] = np.array(data[x])

    return inputdata


def plot_bar(
    names: Union[List[Any], Tuple[Any], np.ndarray],
    counts: Union[List[Any], Tuple[Any], np.ndarray],
    xlabel: str = "x",
    ylabel: str = "y",
    legend: str = "Legend",
    plottitle: str = "Title",
    output_file: str = "Plot",
) -> None:
    """

    Function to create a vertical bar plot

    Parameters
    ----------
    names: List[Any] | Tuple[Any] | np.ndarray
        The numbers or strings for each bar
    counts: List[Any] | Tuple[Any] | np.ndarray
        The height of the bars
    xlabel: str, Default is "x"
        Label of the x-axis
    ylabel: str, Default is "y"
        Label of the y-axis
    legend: str, Default is "Legend"
        Legend for the bar plot
    plottitle: str, Default is "Title"
        The title of the figure
    output_file: str, Default is "Plot"
        The output file location of the .png of the new figure. Do not need the file extension .png

    """

    fig, ax = plt.subplots(figsize=(8, 8))
    chart = ax.bar(names, counts, label=legend)
    ax.bar_label(chart, label_type="center")
    ax.set_xlabel(xlabel, fontsize=12)
    ax.set_ylabel(ylabel, fontsize=12)
    ax.set_title(plottitle, fontsize=12)

    plt.xticks(rotation=90)
    plt.legend()
    plt.tight_layout()

    plt.savefig(f"{output_file}", dpi=400)


def plot_stackedbar(
    names: Union[List[Any], Tuple[Any], np.ndarray],
    bottom_counts: Union[List[Any], Tuple[Any], np.ndarray],
    top_counts: Union[List[Any], Tuple[Any], np.ndarray],
    xlabel: str = "x",
    ylabel: str = "y",
    bottom_legend: str = "Bottom legend",
    top_legend: str = "Top legend",
    plottitle: str = "Title",
    output_file: str = "Plot",
) -> None:
    """

    Function to create a vertically stacked bar plot

    Parameters
    ----------
    names: List[Any] | Tuple[Any] | np.ndarray
        The numbers or strings for each stacked bar
    bottom_counts: List[Any] | Tuple[Any] | np.ndarray
        The height of the bottom bar in the stacked bar plot
    top_counts: List[Any] | Tuple[Any] | np.ndarray
        The height of the top bar in the stacked bar plot
    xlabel: str, Default is "x"
        Label of the x-axis
    ylabel: str, Default is "y"
        Label of the y-axis
    bottom_legend: str, Default is "Bottom legend"
        Legend for the bottom bar in the stacked bar plot
    top_legend: str, Default is "Top legend"
        Legend for the top bar in the stacked bar plot
    plottitle: str, Default is "Title"
        The title of the figure
    output_file: str, Default is "Plot"
        The output file location of the .png of the new figure. Do not need the file extension .png

    """
    fig, ax = plt.subplots(figsize=(8, 8))
    bar_bottom = ax.bar(names, bottom_counts, label=bottom_legend)
    bar_top = ax.bar(names, top_counts, bottom=bottom_counts, label=top_legend)
    ax.bar_label(bar_bottom, label_type="center")
    ax.bar_label(bar_top, label_type="center")
    ax.set_xlabel(xlabel, fontsize=12)
    ax.set_ylabel(ylabel, fontsize=12)
    ax.set_title(plottitle, fontsize=12)

    plt.xticks(rotation=90)
    plt.legend()
    plt.tight_layout()

    plt.savefig(f"{output_file}", dpi=400)


def plot_bar_horizontal(
    names: Union[List[Any], Tuple[Any], np.ndarray],
    counts: Union[List[Any], Tuple[Any], np.ndarray],
    xlabel: str = "x",
    ylabel: str = "y",
    legend: str = "Legend",
    plottitle: str = "Title",
    output_file: str = "Plot",
) -> None:
    """

    Function to create a bar plot

    Parameters
    ----------
    names: List[Any] | Tuple[Any] | np.ndarray
        The numbers or strings for each bar
    counts: List[Any] | Tuple[Any] | np.ndarray
        The width of the bars
    xlabel: str, Default is "x"
        Label of the x-axis
    ylabel: str, Default is "y"
        Label of the y-axis
    legend: str, Default is "Legend"
        Legend for the bar plot
    plottitle: str, Default is "Title"
        The title of the figure
    output_file: str, Default is "Plot"
        The output file location of the .png of the new figure. Do not need the file extension .png

    """
    fig, ax = plt.subplots(figsize=(8, 8))
    chart = ax.barh(names, counts, label=legend)
    ax.bar_label(chart, label_type="center")
    ax.set_xlabel(xlabel, fontsize=12)
    ax.set_ylabel(ylabel, fontsize=12)
    ax.set_title(plottitle, fontsize=12)

    plt.gca().invert_yaxis()
    plt.legend()
    plt.tight_layout()

    plt.savefig(f"{output_file}", dpi=400)


def plot_stackedbar_horizontal(
    names: Union[List[Any], Tuple[Any], np.ndarray],
    left_counts: Union[List[Any], Tuple[Any], np.ndarray],
    right_counts: Union[List[Any], Tuple[Any], np.ndarray],
    xlabel: str = "x",
    ylabel: str = "y",
    left_legend: str = "Left legend",
    right_legend: str = "Right legend",
    plottitle: str = "Title",
    output_file: str = "Plot",
) -> None:
    """

    Function to create a horizontally stacked bar plot

    Parameters
    ----------
    names: List[Any] | Tuple[Any] | np.ndarray
        The numbers or strings for each stacked bar
    left_counts: List[Any] | Tuple[Any] | np.ndarray
        The height of the left bar in the stacked bar plot
    right_counts: List[Any] | Tuple[Any] | np.ndarray
        The height of the right bar in the stacked bar plot
    xlabel: str, Default is "x"
        Label of the x-axis
    ylabel: str, Default is "y"
        Label of the y-axis
    left_legend: str, Default is "Left legend"
        Legend for the bottom bar in the stacked bar plot
    right_legend: str, Default is "Right legend"
        Legend for the top bar in the stacked bar plot
    plottitle: str, Default is "Title"
        The title of the figure
    output_file: str, Default is "Plot"
        The output file location of the .png of the new figure. Do not need the file extension .png

    """
    fig, ax = plt.subplots(figsize=(8, 8))
    bar_left = ax.barh(names, left_counts, label=left_legend)
    bar_right = ax.barh(names, right_counts, left=left_counts, label=right_legend)
    ax.bar_label(bar_left, label_type="center")
    ax.bar_label(bar_right, label_type="center")
    ax.set_xlabel(xlabel, fontsize=12)
    ax.set_ylabel(ylabel, fontsize=12)
    ax.set_title(plottitle, fontsize=12)

    plt.gca().invert_yaxis()
    plt.legend()
    plt.tight_layout()

    plt.savefig(f"{output_file}", dpi=400)



from typing import Dict, List, Union

import numpy as np
from pymoo.core.problem import Problem

from nemo_bo.opt.constraints import ConstraintsList
from nemo_bo.opt.objectives import ObjectivesList
from nemo_bo.opt.variables import VariablesList


class NSGAProblemWrapper(Problem):
    """

    Class for defining a problem for the minimise function in NSGA-based methods

    Parameters
    ----------
    xl: int | float | List[int | float]
        Lower bounds for the variables for the NSGA algorithm. If integer, all lower bounds are equal to the integer
        value
    xu: int | float | List[int | float]
        Upper bounds for the variables for the NSGA algorithm. If integer, all upper bounds are equal to the integer
        value
    variables: VariablesList
        VariablesList object that contains information about all variables
    objectives: ObjectivesList
        ObjectivesList object that contains information about all objectives
    constraints: ConstraintsList
        ConstraintsList object that contains information about all constraints

    """

    def __init__(
        self,
        xl: Union[int, float, List[Union[int, float]]],
        xu: Union[int, float, List[Union[int, float]]],
        variables: VariablesList,
        objectives: ObjectivesList,
        constraints: ConstraintsList,
    ):
        self.variables = variables
        self.objectives = objectives
        self.constraints = constraints

        n_constr = self.constraints.n_constr if constraints is not None else 0

        super().__init__(
            n_var=self.variables.n_var,
            n_obj=self.objectives.n_obj,
            n_constr=n_constr,
            xl=xl,
            xu=xu,
        )

    def _evaluate(self, X: np.ndarray, out: Dict[str, np.ndarray], **kwargs) -> None:
        """

        Function to calculate the objective(s) and constraint(s) (if any) values

        Parameters
        ----------
        X: np.ndarray
            NSGA generated candidates' X-values, which are min-max normalised
        out: Dict[str, np.ndarray]
            Contains the objective(s) and constraint(s) (if any) values for the generated candidates

        """
        # Un-transform the min-max normalised X array
        X = (X * (np.array(self.variables.upper_bounds) - np.array(self.variables.lower_bounds))) + np.array(
            self.variables.lower_bounds
        )

        # Infer the objective values by drawing samples from the distribution at the X-values
        if self.variables.num_cat_descriptor_var > 0:
            F = self.objectives.draw_samples(self.variables.descriptor_to_name(X))
        else:
            F = self.objectives.draw_samples(X)

        # * -1 for objectives that are maximising
        for obj_index, obj in enumerate(self.objectives.max_bool_dict):
            if self.objectives.max_bool_dict[obj]:
                F[:, obj_index] *= -1
        out["F"] = F

        # Evaluate the pymoo constraints
        if self.n_constr > 0:
            # Min-max normalises the X array
            X = (X - np.array(self.variables.lower_bounds)) / (
                np.array(self.variables.upper_bounds) - np.array(self.variables.lower_bounds)
            )
            G = np.zeros((X.shape[0], self.n_constr))
            for index, constr in enumerate(self.constraints.constraints):
                G[:, index] = constr.evaluate_pymoo_constraint(X)
            out["G"] = G



# The NSGA-based methods were copied and adapted from J. Blank and K. Deb, pymoo: Multi-Objective Optimization in
# Python, in IEEE Access, vol. 8, pp. 89497-89509, 2020, doi: 10.1109/ACCESS.2020.2990567
# https://ieeexplore.ieee.org/document/9078759
# https://github.com/anyoptimization/pymoo
import numpy as np

from pymoo.operators.selection.tournament import compare, TournamentSelection

from nemo_bo.acquisition_functions.nsga_improvement.nsga3 import NSGA3


# =========================================================================================================
# Implementation
# =========================================================================================================


def comp_by_rank_and_ref_line_dist(pop, P, **kwargs):
    S = np.full(P.shape[0], np.nan)

    for i in range(P.shape[0]):
        a, b = P[i, 0], P[i, 1]

        # if at least one solution is infeasible
        if pop[a].CV > 0.0 or pop[b].CV > 0.0:
            S[i] = compare(a, pop[a].CV, b, pop[b].CV, method="smaller_is_better", return_random_if_equal=True)

        # both solutions are feasible
        else:

            # if in the same niche select by rank
            if pop[a].get("niche") == pop[b].get("niche"):

                if pop[a].get("rank") != pop[b].get("rank"):
                    S[i] = compare(a, pop[a].get("rank"), b, pop[b].get("rank"), method="smaller_is_better")

                else:
                    S[i] = compare(
                        a, pop[a].get("dist_to_niche"), b, pop[b].get("dist_to_niche"), method="smaller_is_better"
                    )

        if np.isnan(S[i]):
            S[i] = np.random.choice([a, b])

    return S[:, None].astype(int)


class UNSGA3(NSGA3):
    def __init__(self, ref_dirs, selection=TournamentSelection(func_comp=comp_by_rank_and_ref_line_dist), **kwargs):
        super().__init__(ref_dirs, selection=selection, **kwargs)
